#!/usr/bin/env python
#

import os, hashlib
import sys
import time
import pandas
import requests
import warnings
import numpy as np

LOGIN_URL = "https://irsa.ipac.caltech.edu/account/signon/login.do"

import base64

from configparser import ConfigParser
from astropy.io import fits

from .utils.tools import is_running_from_notebook


_SOURCEDIR = os.path.dirname(os.path.realpath(__file__))


_ENCRYPT_FILE = os.path.expanduser("~") + "/.ztfquery"
_ENCRYPTING_FILE = os.path.expanduser("~") + "/.queryirsa"


LOCALSOURCE = os.getenv("ZTFDATA", "./Data/")
CCIN2P3_SOURCE = "/sps/ztf/data/"


# ================= #
#  High level tools #
# ================= #
def get_file(
    filename,
    suffix=None,
    downloadit=True,
    verbose=False,
    check_suffix=True,
    dlfrom="irsa",
    overwrite=False,
    maxnprocess=4,
    exist=True,
    test_file=True,
    squeeze=True,
    show_progress=True,
    client=None,
    wait=None,
    fill_notexist="None",
    **kwargs,
):
    """Get full path associate to the filename.
    If you don't have it on your computer, this downloads it for you.

    Parameters
    ----------
    filename: [string]
        name of the file you want.
        raw, cal and sci filenames are accepted.

    suffix: [string] -optional-
        actual suffix of the file you want. By default it is that of the filename
        but you can request associated files with other suffix.
        For instance:
        if filename = ztf_20190917468333_000698_zi_c03_o_q2_sciimg.fits and suffix='mskimg.fits'
        you will be looking for ztf_20190917468333_000698_zi_c03_o_q2_mskimg.fits.
        - If you input a raw file, then suffix is used as 'imgtypecode'

    downloadit: [bool] -optional-
        If you do not have the file locally, shall this download it for you ?

    wait: [None/string/float] -optional-
        Waiting time for dwnloaded the file. It helps mostly when dask is massively multi-downloading.
        - None: if Client, wait -> len(to_be_downloaded)/100 (so 100 per second) ; corresponds to wait="100"
        - "None": no limit
        - "float_in_string": waiting time reported to the size of file to download: len(to_be_downloaded)/wait
        - float: absolute time to wait.

    **kwargs goes to download_from_filename

    Returns
    -------
    fullpath (or None if not data)

    """
    from .buildurl import filename_to_url

    local_filenames = np.asarray(
        [
            filename_to_url(
                filename_,
                suffix=suffix_,
                verbose=verbose,
                source="local",
                check_suffix=check_suffix,
            )
            for filename_ in np.atleast_1d(filename)
            for suffix_ in np.atleast_1d(suffix)
        ]
    )
    if not exist:
        return local_filenames

    # local_filenames
    if overwrite:
        flag_todl = np.asarray(np.ones(len(local_filenames)), dtype="bool")
    else:
        flag_todl = np.asarray(
            [
                (not os.path.isfile(f_))
                or (test_file and ".fits" in f_ and _is_fitsfile_bad_(f_))
                for f_ in local_filenames
            ]
        )

    # DL if needed (and wanted)
    if np.any(flag_todl) and downloadit:
        if client is not None and wait is None:
            wait = "100"
        if type(wait) is str:
            if wait == "None":
                wait = None
            else:
                try:
                    wait = len(local_filenames[flag_todl]) / float(wait)
                except:
                    warnings.warn(
                        f"cannot parse the inpout waiting time {wait} -> None used."
                    )
                    wait = None

        f_ = download_from_filename(
            local_filenames[flag_todl],
            show_progress=show_progress,
            host=dlfrom,
            overwrite=True,
            maxnprocess=maxnprocess,
            client=client,
            wait=wait,
            check_suffix=check_suffix,
        )
        if client is not None:
            from dask.distributed import wait as dwait

            _ = dwait(f_)

    # - Output
    if fill_notexist == "remove":
        local_filenames = [f for f in local_filenames if os.path.isfile(f)]

    elif fill_notexist != "None":
        local_filenames = [
            f if os.path.isfile(f) else fill_notexist for f in local_filenames
        ]

    if len(local_filenames) == 1 and squeeze:
        return local_filenames[0]

    return local_filenames


def filefracday_to_local_rawdata(filefracday, ccdid="*"):
    """ """
    from glob import glob

    rawfilename = filefracday_and_ccdid_to_rawfilename(filefracday, ccdid=ccdid)
    return np.sort(glob(rawfilename))


def filefracday_and_ccdid_to_rawfilename(filefracday, ccdid, source="local"):
    from .buildurl import filefrac_to_year_monthday_fracday, _source_to_location_

    filefracday = str(filefracday)
    year, month, day, fracday = filefrac_to_year_monthday_fracday(filefracday)
    source = _source_to_location_(source)
    cstring = "*" if ccdid in ["*", "all"] else f"_c{ccdid:02d}_"
    return os.path.join(
        source,
        "raw",
        year,
        f"{month}{day}",
        fracday,
        f"ztf_{filefracday}*{cstring}*.fits.fz",
    )


def bulk_get_file(filenames, client=None, suffix=None, as_dask="delayed", **kwargs):
    """
    Parameters
    ----------
    as_dask: [string] -optional-
         could be
         - delayed
         - futures
         - gathered
    """
    import dask

    if client is None and as_dask in ["gather", "gathered"]:
        as_dask = "compute"
    d_files = [
        dask.delayed(get_file)(
            filename, suffix=suffix, show_progress=False, maxnprocess=1, **kwargs
        )
        for filename in filenames
    ]
    if as_dask == "delayed":
        return d_files
    if as_dask in ["computed", "compute"]:
        return dask.delayed(list)(d_files).compute()

    if as_dask == "persist":
        if client is not None:
            return client.persist(d_files)
        return [f_.persist() for f_ in d_files]

    futures = client.compute(d_files)
    if as_dask == "futures":
        return futures
    if as_dask in ["gather", "gathered"]:
        return client.gather(futures)
    raise ValueError(f"Cannot parse the given as_dask {as_dask}")


def get_filedataframe(filenames):
    """get a dataframe of the files"""
    import pandas

    fileserie = pandas.Series(filenames, name="filename")

    fdata = pandas.DataFrame.from_records(fileserie.apply(parse_filename))
    fdata["isfile"] = fileserie.apply(os.path.isfile)
    merged = fdata.merge(fileserie, left_index=True, right_index=True)
    return merged


def parse_filename(filename, as_serie=True):
    """ """
    from .buildurl import parse_filename

    if as_serie:
        return pandas.Series(parse_filename(filename))
    return parse_filename(filename)


def download_from_filename(
    filename,
    suffix=None,
    verbose=False,
    overwrite=False,
    auth=None,
    nodl=False,
    host="irsa",
    maxnprocess=4,
    show_progress=True,
    check_suffix=True,
    client=None,
    wait=None,
    **kwargs,
):
    """Download the file associated to the given filename"""
    if host not in ["irsa", "ccin2p3"]:
        raise ValueError(f"Only 'irsa' and 'ccin2p3' host implemented: {host} given")

    from .buildurl import filename_to_url

    if auth is None:
        auth = _load_id_(host)

    remote_filename = []
    local_filename = []
    for file_ in np.atleast_1d(filename):
        remote_filename.append(
            filename_to_url(
                file_,
                suffix=suffix,
                verbose=verbose,
                source=host,
                check_suffix=check_suffix,
            )
        )
        local_filename.append(
            filename_to_url(
                file_,
                suffix=suffix,
                verbose=verbose,
                source="local",
                check_suffix=check_suffix,
            )
        )

    if nodl:
        return [remote_filename, local_filename]

    if host == "ccin2p3":
        return [
            CCIN2P3.scp(remote_filename_, local_filename_, auth=auth)
            for remote_filename_, local_filename_ in zip(
                remote_filename, local_filename
            )
        ]

    else:
        nprocess = np.min([maxnprocess, len(local_filename)])

        return download_url(
            remote_filename,
            local_filename,
            nprocess=nprocess,
            client=client,
            wait=wait,
            overwrite=overwrite,
            verbose=verbose,
            cookies=get_cookie(*auth),
            show_progress=show_progress,
            **kwargs,
        )


def _parse_filename_(filename, builddir=False, squeeze=True, exists=False):
    """ """
    print("IS THIS USED ?")
    from glob import glob

    directory = os.path.dirname(filename)
    # basename  = os.path.basename(filename).split(".")[0]
    # extension = filename.split(".")[-1]

    if builddir:
        oldmask = os.umask(0o002)
        os.makedirs(directory, exist_ok=True)

    # unique object
    if "*" in filename and not exists:
        warnings.warn(
            "apparent variable conflict, filename contains '*', but exists is False."
        )

    localfile = glob(filename) if exists else np.atleast_1d(filename)
    if squeeze:
        if len(localfile) == 0:
            return None
        if len(localfile) == 1:
            return localfile[0]

    return localfile


# ================= #
#  Crypting         #
# ================= #


def _load_id_(which, askit=True, token_based=False):
    """returns login information for the requested enty"""
    import base64

    config = ConfigParser()
    config.read(_ENCRYPT_FILE)

    if which in ["fritz"]:
        token_based = True

    if which not in config.sections():
        if not askit:
            raise AttributeError(
                f"No {which} account setup. Add then in .ztfquery or run ztfquery.io.set_account({which})"
            )
        else:
            warnings.warn(f"No {which} account setup, please provide it")
            set_account(which, token_based=token_based)
            config = ConfigParser()
            config.read(_ENCRYPT_FILE)

    if not token_based:
        return config[which.lower()]["username"], base64.b64decode(
            config[which.lower()]["password"][2:-1]
        ).decode("utf-8")

    return base64.b64decode(config[which.lower()]["token"][2:-1]).decode("utf-8")


def set_account(
    which,
    username=None,
    password=None,
    token=None,
    test=True,
    force=False,
    token_based=False,
    no_user=False
    ):
    """Setup the username and password (simply encrypted!) for the given `which` account.
    Saved in ~/.ztfquery
    """
    import base64
    import getpass

    config = ConfigParser()
    config.read(_ENCRYPT_FILE)

    if which in ["fritz"]:
        token_based = True

    if which in ["logs"]:
        no_user = True

    if token_based:
        if token is None:
            token = input(f"Enter your {which} token:")
    else:
        # - Name & Password
        if username is None:
            if no_user:
                username = "None"
            else:
                username = input(f"Enter your {which} login: ")

        if password is None:
            password = getpass.getpass()

    #
    # -> Starting tests
    if test:
        wrong_ = False
        if which == "irsa":
            if not test_irsa_account([username, password]):
                warnings.warn(
                    "The irsa_test for you account returns False. Most likely you provided incorrect logins"
                )
                wrong_ = True
        else:
            if not token_based:
                warnings.warn(
                    f"No test designed for {which}. Cannot test if logins are correct."
                )
            else:
                warnings.warn(
                    f"No test designed for {which}. Cannot test if token is correct."
                )

        if wrong_ and not force:
            if not token_based:
                raise ValueError(
                    f"Bad username/passworg for {which}. force=False so the logins are not stored. "
                )
            else:
                raise ValueError(
                    "Bad token for {which}. force=False so the logins are not stored."
                )
    # <- end of tests
    #
    if not token_based:
        password_ = base64.b64encode(password.encode("utf-8"))
        config[which.lower()] = {"username": username, "password": password_}
    else:
        token_ = base64.b64encode(token.encode("utf-8"))
        config[which.lower()] = {"token": token_}

    with open(_ENCRYPT_FILE, "w") as configfile:
        config.write(configfile)


#
# TEST
#
# - Password testing
def test_irsa_account(auth=None):
    """returns True if the IRSA account is correctly set."""
    if auth is None:
        auth = _load_id_("irsa")
    return ".ipac.caltech.edu" in get_cookie(*auth)._cookies


# - File testing
def get_localfiles(extension="*", startpath=None):
    """Look for all file with the given extension recursively starting from `startpath`.
    (based on glob)

    Parameters
    ----------
    extension: [string] -optional-
        All the 'file.{}'.format(extension) will be looked at.
        (first '.' ignored such that extension='.fits' is equivalent to extension='fits')

    startpath: [None or path] -optional-
        From which directory does this start to look at.
        If None: $ZTFDATA (stored as io.LOCALSOURCE) will be used.

    Returns
    -------
    list of file.
    """
    from glob import glob

    if startpath is None:
        startpath = LOCALSOURCE
    if extension.startswith("."):
        extension = extension[1:]

    return [f for f in glob(startpath + f"**/*.{extension}", recursive=True)]


def run_full_filecheck(
    extension="*",
    startpath=None,
    erasebad=True,
    redownload=False,
    nprocess=4,
    show_progress=True,
    **kwargs,
):
    """Look for all file with the given extension recursively starting from `startpath` and checks if the file
    is usable ok not. This returns the bad files.

    Parameters
    ----------

    // Data files
    extension: [string] -optional-
        All the 'file.{}'.format(extension) will be looked at.
        (first '.' ignored such that extension='.fits' is equivalent to extension='fits')

    startpath: [None or path] -optional-
        From which directory does this start to look at.
        If None: $ZTFDATA (stored as io.LOCALSOURCE) will be used.

    // Check options

    erasebad: [bool] -optional-
        Do you want to remove from your local directory the corrupted files ?

    redownload: [bool] -optional-
        Shall corrupted file be automatically re downloaded ?
        (Only works for IRSA files ('/sci/','/raw/', '/ref/', '/cal/')

    nprocess: [int] -optional-
        Number of paralell processing

    show_progress: [bool] -optional-
        Do you want to show the progress bar ?

    Returns
    -------
    list of corrupted/bad files (might already be removed, see erasebad)

    """
    all_ztffiles = get_localfiles(extension=extension, startpath=startpath)
    print(f"{len(all_ztffiles)} files to check")
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        badfiles = test_files(
            all_ztffiles,
            erasebad=erasebad,
            nprocess=nprocess,
            show_progress=show_progress,
            redownload=redownload,
            **kwargs,
        )

    return badfiles


def test_files(
    filename,
    erasebad=True,
    nprocess=1,
    show_progress=True,
    redownload=False,
    verbose=False,
    **kwargs,
):
    """

    Parameters
    ----------
    filename: [fiulepath or list of]
        File(s) to be checked.

    erasebad: [bool] -optional-
        Do you want to remove from your local directory the corrupted files ?

    redownload: [bool] -optional-
        Shall corrupted file be automatically re downloaded ?
        (Only works for IRSA files ('/sci/','/raw/', '/ref/', '/cal/')

    nprocess: [int] -optional-
        Number of paralell processing

    show_progress: [bool] -optional-
        Do you want to show the progress bar ?

    Returns
    -------
    list of corrupted/bad files (might already be removed, see erasebad)

    """
    if nprocess is None:
        nprocess = 1
    elif nprocess < 1:
        raise ValueError("nprocess must 1 or higher (None means 1)")

    filename = np.atleast_1d(filename)

    if nprocess == 1:
        fileissue = [
            f
            for f in filename
            if not _test_file_(f, erasebad=erasebad, redownload=redownload, **kwargs)
        ]
    else:
        import multiprocessing

        if show_progress:
            from astropy.utils.console import ProgressBar

            bar = ProgressBar(len(filename), ipython_widget=is_running_from_notebook())
        else:
            bar = None

        erasebad_ = [erasebad] * len(filename)
        fileissue = []

        with multiprocessing.Pool(nprocess) as p:
            # Da Loop
            for j, isgood in enumerate(
                p.imap(_test_file_multiprocess_, zip(filename, erasebad_))
            ):
                if bar is not None:
                    bar.update(j)
                if not isgood:
                    fileissue.append(filename[j])

            if bar is not None:
                bar.update(len(filename))

    if len(fileissue) > 0:
        warnings.warn(f"{len(fileissue)} file failed")
        if redownload:
            from .buildurl import _localsource_to_source_

            to_download_urls, locations = np.asarray(
                [_localsource_to_source_(filename) for filename in fileissue]
            ).T
            source_to_dl = ["irsa"]
            for source in source_to_dl:
                source_dl = np.in1d(locations, [source])
                print(f"Downloading {len(source_dl[source_dl])} files from {source}")
                download_url(
                    np.asarray(to_download_urls)[source_dl],
                    np.asarray(fileissue)[source_dl],
                    show_progress=show_progress,
                    verbose=verbose,
                    overwrite=True,
                    nprocess=nprocess,
                    cookies=get_cookie(*_load_id_(source)),
                    **kwargs,
                )
            for source_ in np.unique(locations):
                if source_ is not None and source_ not in source_to_dl:
                    warnings.warn(
                        f"files from {source_} have not downloaded (not implemented)."
                    )

        return fileissue


def _test_file_multiprocess_(args):
    """ """
    filename, erasebad = args
    return _test_file_(filename, erasebad=erasebad, fromdl=False, redownload=False)


def _are_fitsfiles_bad_(filenames, test_exist=True):
    """ """
    return [_is_fitsfile_bad_(f_, test_exist=test_exist) for f_ in filenames]


def _is_fitsfile_bad_(filename, test_exist=True):
    """ """
    if not os.path.isfile(filename):
        return test_exist

    try:
        _ = fits.getdata(filename)
        return False
    except:
        return True


def _is_textfile_bad_(filename):
    """ """
    try:
        _ = open(filename).read().splitlines()
        return False
    except:
        return True


def _test_file_(filename, erasebad=True, fromdl=False, redownload=False, verbose=True):
    """ """
    propissue = dict(
        erasebad=erasebad, fromdl=fromdl, redownload=redownload, verbose=verbose
    )
    # Fits file
    if ".fits" in filename:
        if not hash_for_file_exists(filename):
            try:
                _ = fits.getdata(filename)
                calculate_and_write_hash(filename)
            except FileNotFoundError:
                warnings.warn(f"[Errno 2] No such file or directory: {filename}")
            except:
                _fileissue_(filename, **propissue)
                return False

    # txt file
    elif ".txt" in filename:
        if not hash_for_file_exists(filename):
            try:
                _ = open(filename).read().splitlines()
                calculate_and_write_hash(filename)
            except FileNotFoundError:
                warnings.warn(f"[Errno 2] No such file or directory: {filename}")
            except:
                _fileissue_(filename, **propissue)
                return False

    # other extensions
    else:
        warnings.warn("no file testing made for .%s files" % filename.split(".")[-1])

    return True


def _fileissue_(filename, erasebad=True, fromdl=False, redownload=False, verbose=True):
    """ """
    if fromdl:
        warnings.warn(f"Download failed {filename} seems corrupted (cannot open)")
    else:
        warnings.warn(f"cannot open file {filename}")

    if erasebad:
        warnings.warn(f"removing {filename}")
        os.remove(filename)
    else:
        warnings.warn(f"{filename} NOT ERASED")

    if redownload:
        from .buildurl import _localsource_to_source_

        url_to_dl, location = _localsource_to_source_(filename)
        if url_to_dl is not None:
            download_single_url(
                url_to_dl,
                fileout=filename,
                overwrite=True,
                verbose=verbose,
                cookies=get_cookie(*_load_id_(location)),
            )
        else:
            warnings.warn("No url to donwload, redownload ignored")


# ================= #
#   Logging Tools   #
# ================= #
def get_cookie(username, password):
    """Get a cookie from the IPAC login service

    Parameters
    ----------
    username: [str]
        The IPAC account username
    password: [str]
        The IPAC account password
    """
    url = "%s?josso_cmd=login&josso_username=%s&josso_password=%s" % (
        LOGIN_URL,
        username,
        password,
    )
    return requests.get(url).cookies


def _download_(args):
    """To be used within _ZTFDownloader_.download_data()
    url, fileout,overwrite,verbose = args
    """
    url, fileout, overwrite, verbose, wait = args
    download_single_url(
        url, fileout=fileout, overwrite=overwrite, verbose=verbose, wait=wait
    )


def download_url(
    to_download_urls,
    download_location,
    show_progress=True,
    verbose=True,
    wait=None,
    overwrite=False,
    nprocess=None,
    cookies=None,
    client=None,
    **kwargs,
):
    """ """
    #
    # - Dask Client
    if client is not None:
        from dask import delayed

        d_download = [
            delayed(download_single_url)(
                url,
                fileout=fileout,
                show_progress=False,
                overwrite=overwrite,
                verbose=False,
                wait=wait,
                cookies=cookies,
                **kwargs,
            )
            for url, fileout in zip(to_download_urls, download_location)
        ]
        return client.compute(d_download)

    #
    # - MultiProcessing (or not)
    if nprocess is None:
        nprocess = 1
    elif nprocess < 1:
        raise ValueError("nprocess must 1 or higher (None means 1)")

    if nprocess == 1:
        # Single processing
        if verbose:
            warnings.warn("No parallel downloading")
        for url, fileout in zip(to_download_urls, download_location):
            download_single_url(
                url,
                fileout=fileout,
                show_progress=show_progress,
                overwrite=overwrite,
                verbose=verbose,
                cookies=cookies,
                wait=wait,
                **kwargs,
            )
    else:
        # Multi processing
        import multiprocessing

        if show_progress:
            from astropy.utils.console import ProgressBar

            bar = ProgressBar(
                len(to_download_urls), ipython_widget=is_running_from_notebook()
            )
        else:
            bar = None

        if verbose:
            warnings.warn("parallel downloading ; asking for %d processes" % nprocess)

        # Passing arguments
        overwrite_ = [overwrite] * len(to_download_urls)
        verbose_ = [verbose] * len(to_download_urls)
        wait_ = [wait] * len(to_download_urls)
        with multiprocessing.Pool(nprocess) as p:
            # Da Loop
            for j, result in enumerate(
                p.imap_unordered(
                    _download_,
                    zip(
                        to_download_urls, download_location, overwrite_, verbose_, wait_
                    ),
                )
            ):
                if bar is not None:
                    bar.update(j)

            if bar is not None:
                bar.update(len(to_download_urls))


def download_single_url(
    url,
    fileout=None,
    overwrite=False,
    verbose=True,
    cookies=None,
    show_progress=True,
    chunk=1024,
    wait=None,
    randomize_wait=True,
    filecheck=True,
    erasebad=True,
    **kwargs,
):
    """Download the url target using requests.get.
    the data is returned (if fileout is None) or stored in `fileout`
    """
    if wait is not None:
        waiting = wait if not randomize_wait else np.random.uniform(0, wait)
        time.sleep(waiting)

    if fileout is not None and not overwrite and os.path.isfile(fileout):
        if verbose:
            warnings.warn("%s already exists: skipped" % fileout)
        return
    else:
        if verbose and fileout:
            warnings.warn("downloading %s to %s" % (url, fileout))

    # = Password and Username
    if cookies is None:
        cookies = get_cookie(*_load_id_("irsa"))

    # - requests options
    download_prop = dict(cookies=cookies, stream=True)
    for k, v in kwargs.items():
        download_prop[k] = v

    if cookies in ["no_cookies"]:
        _ = download_prop.pop("cookies")

    request_fnc = "get" if not "data" in download_prop else "post"
    # = Where should the data be saved?
    if fileout is not None:
        directory = os.path.dirname(fileout)
        oldmask = os.umask(0o002)

        if not os.path.exists(directory):
            os.makedirs(directory, exist_ok=True)

    else:
        download_prop["stream"] = False
        return getattr(requests, request_fnc)(url, **download_prop)

    # With Progress bar?
    if not show_progress:
        response = getattr(requests, request_fnc)(url, **download_prop)
        if response.status_code == 200:
            with open(fileout, "wb") as f:
                for data in response.iter_content(chunk):
                    f.write(data)

    else:
        from astropy.utils.console import ProgressBar

        response = getattr(requests, request_fnc)(url, **download_prop)
        if response.status_code == 200:
            chunk_barstep = 500
            f = open(fileout, "wb")
            with ProgressBar(
                int(response.headers.get("content-length")) / (chunk_barstep * chunk),
                ipython_widget=is_running_from_notebook(),
            ) as bar:
                for i, data in enumerate(response.iter_content(chunk_size=chunk)):
                    if i % chunk_barstep == 0:
                        bar.update()
                    f.write(data)
            f.close()
            calculate_and_write_hash(fileout)

    if filecheck:
        _test_file_(fileout, erasebad=erasebad, fromdl=True)


# ============== #
#                #
#  CC-IN2P3      #
#                #
# ============== #
class CCIN2P3(object):
    """ """

    def __init__(self, auth=None, connect=True):
        """ """
        if self.running_at_cc:
            self._connected = True
        else:
            self.load_ssh(auth=auth)
            self._connected = False
            if connect:
                self.connect(auth=auth)

    def load_ssh(self, auth=None):
        from paramiko import SSHClient

        self._auth = auth
        self._ssh = SSHClient()
        self._ssh.load_system_host_keys()

    def connect(self, auth=None):
        """ """
        if auth is None:
            auth = _load_id_("ccin2p3")

        username, password = auth
        try:
            self._ssh.connect("cca.in2p3.fr", username=username, password=password)
        except:
            raise IOError("Cannot connect to cca.in2p3.fr with given authentification")

        self._connected = True

    @classmethod
    def scp(cls, fromfile, tofile, auth=None):
        """ """
        if fromfile.startswith(CCIN2P3_SOURCE):
            method = "scp_get"
        elif tofile.startswith(CCIN2P3_SOURCE):
            method = "scp_put"
        else:
            raise ValueError(
                f"None of fromfile or tofile stars with {CCIN2P3_SOURCE}. Cannot use scp(), see scp_get or scp_put"
            )

        this = cls(auth=auth, connect=True)
        return getattr(this, method)(fromfile, tofile)

    def scp_get(self, remotefile, localfile, auth=None, overwrite=False):
        """ """
        from scp import SCPClient

        if not self._connected:
            self.connect(auth)

        directory = os.path.dirname(localfile)
        oldmask = os.umask(0o002)

        if not os.path.exists(directory):
            warnings.warn(f"scp_get(): creating {directory}")

            os.makedirs(directory, exist_ok=True)

        with SCPClient(self.ssh.get_transport()) as scp:
            scp.get(remotefile, localfile)

    def scp_put(self, localfile, remotefile, auth=None):
        """ """
        from scp import SCPClient

        if not self._connected:
            self.connect(auth)

        with SCPClient(self.ssh.get_transport()) as scp:
            scp.put(localfile, remotefile)

    def query_catalog(self, ra, dec, radius, catname="gaia", depth=7, **kwargs):
        """query catalog ; works only when logged at the CCIN2P3"""
        if not self.running_at_cc:
            raise IOError("Only works if running from the ccin2p3")

        import os
        from htmcatalog import htmquery

        LSST_REFCAT_DIR = "/sps/lsst/datasets/refcats/htm/v1"
        KNOW_REFCAT = {
            "gaiadr1": "gaia_DR1_v1",
            "gaiadr2": "gaia_dr2_20190808",
            "ps1dr1": "ps1_pv3_3pi_20170110",
            "sdssdr9": "sdss-dr9-fink-v5b",
        }
        KNOW_REFCAT["gaia"] = KNOW_REFCAT["gaiadr2"]
        KNOW_REFCAT["ps1"] = KNOW_REFCAT["ps1dr1"]
        if catname not in KNOW_REFCAT:
            raise ValueError(
                f"unknown catalog {catname}. Aviability: "
                + ", ".join(list(KNOW_REFCAT.keys()))
            )

        hq = htmquery.HTMQuery(
            depth, os.path.join(LSST_REFCAT_DIR, KNOW_REFCAT[catname])
        )
        return hq.fetch_cat(ra, dec, radius, **kwargs)

    # ============= #
    #  Properties   #
    # ============= #
    @property
    def ssh(self):
        """ """
        return self._ssh

    @property
    def running_at_cc(self):
        """ """
        hostname = os.uname()[1]
        return "cca" in hostname or "ccwige" in hostname


# =============== #
#                 #
#  HASH tools     #
#                 #
# =============== #


def calculate_hash(fname):
    """ """
    f = open(fname, "rb")
    hash_md5 = hashlib.md5()
    for chunk in iter(lambda: f.read(4096), b""):
        hash_md5.update(chunk)
    hexdigest = hash_md5.hexdigest()
    f.close()
    return hexdigest


def calculate_and_write_hash(fname):
    """ """
    f = open(fname, "rb")
    hash_md5 = hashlib.md5()
    for chunk in iter(lambda: f.read(4096), b""):
        hash_md5.update(chunk)
    hexdigest = hash_md5.hexdigest()
    f.close()
    f_hash = open(f"{fname}.md5", "w")
    f_hash.write(hexdigest)
    f_hash.close()


def read_hash(fname):
    """ """
    f_hash = open(f"{fname}.md5", "r")
    hash_md5 = f_hash.read()
    return hash_md5


def compare_hash(fname):
    """ """
    f_hash = open(hash_fname, "r")
    hash_md5_read = f_hash.read()
    hash_md5_calculated = calculate_hash(fname)
    if hash_md5_read == hash_md5_calculated:
        return True
    else:
        return False


def hash_for_file_exists(fname):
    """ """
    return os.path.exists(f"{fname}.md5")



#!/usr/bin/env python
#

""" Library containing field information """
import os, sys
import pandas
import numpy as np
import warnings

from astropy import units, coordinates, time
import matplotlib.pyplot as mpl
from matplotlib.patches import Polygon


from .io import _SOURCEDIR

_FIELD_SOURCE = os.path.join(_SOURCEDIR, "data/ztf_fields.txt")
FIELD_DATAFRAME = pandas.read_csv(_FIELD_SOURCE, index_col="ID")
FIELDSNAMES = FIELD_DATAFRAME.index.values


_CCD_COORDS  = pandas.read_csv( os.path.join(_SOURCEDIR, "data/ztf_ccd_layout.tbl")).rename(columns={"CCD ":"CCD"}) # corner of each CCDS
_QUAD_COORDS  = pandas.read_csv( os.path.join(_SOURCEDIR,"data/ztf_ccd_quad_layout.tbl"))#.rename(columns={"CCD ":"CCD"}) # corner of each CCDS

FIELD_COLOR = {1: "C2", 2: "C3", 3:"C1"}
FIELD_CMAP = {1: mpl.cm.Greens, 2:mpl.cm.Reds, 3:mpl.cm.Oranges}
FIELDNAME_COLOR = {"zg": "C2", "zr":"C3", "zi":"C1"}

_PLOTORIGIN = 180

def _load_fields_geoserie_(inclquad=False, inclccd=False):
    """ Loads the FIELDS_GEOSERIE global variable 
    = Internal Tools =
    """
    if inclquad:
        global FIELD_QUADS_GEOSERIE
    elif inclccd:
        global FIELD_CCDS_GEOSERIE
    else:
        global FIELDS_GEOSERIE
    
    try:
        from geopandas import geoseries
    except ImportError:
        warnings.warn("You do not have geopandas, Please run pip install geopandas.")
        if not inclccd:
            FIELDS_GEOSERIE = None
        else:
            FIELD_CCDS_GEOSERIE = None 
        return
    
    field_verts = get_field_vertices(fieldid=FIELDSNAMES, inclquad=inclquad, inclccd=inclccd, as_dict=True, as_polygon=True)
    if inclquad:
        FIELD_QUADS_GEOSERIE = geoseries.GeoSeries(field_verts)
    elif inclccd:
        FIELD_CCDS_GEOSERIE = geoseries.GeoSeries(field_verts)
    else:
        FIELDS_GEOSERIE = geoseries.GeoSeries(field_verts)

    
def get_fields_geoserie(inclquad=False, inclccd=False):
    """ returns the global variable FIELDS_GEOSERIE, creates it if necessary """
    if inclquad:
        if not hasattr(sys.modules[__name__], 'FIELD_QUADS_GEOSERIE'):
            _load_fields_geoserie_(inclccd=False, inclquad=True)
        
        return FIELD_QUADS_GEOSERIE
    
    elif inclccd:
        if not hasattr(sys.modules[__name__], 'FIELD_CCDS_GEOSERIE'):
            _load_fields_geoserie_(inclccd=True)
        
        return FIELD_CCDS_GEOSERIE
    
    else:
        if not hasattr(sys.modules[__name__], 'FIELDS_GEOSERIE'):
            _load_fields_geoserie_(inclccd=False)
        
        return FIELDS_GEOSERIE
        

# ------------------ #
#                    #
# Generic Tools      #
#                    #
# ------------------ #
def get_fieldid(grid=None, decrange=None, rarange=None, 
                gallrange=None, galbrange=None, 
                ecllrange=None, eclbrange=None, 
                ebvrange=None, verbose=False):
                
    """ 

    Parameters
    ----------
    grid: [None/str] -optional-
        Select the grid you want. If None the both will be considered.
        'main' or 'secondary' expected otherwise.

    decrange, rarange, galbrange, gallrange, ecllgrange, eclbrange: [2-array or None] -optional-
        dec, ra, galactic l, galactic b, ecliptic long and ecliptic lat to be considered [inclusive]. 
        In degree.
        3 formats available:
        - None: means no selection
        - [min,max]: means range to be considered, None means no limit. 
          example: decrange=[-10,None] means dec>-10.
        - [[min1,max1],[min2,max2], ...]: means zones to be considered.
          example: decrange=[[None, -10],[5,None]] will simply exclude the [-10,5] dec range.
        - None means no selection.

    ebvrange: [2-array or None] -optional-
        same format as 'decrange' but for Mily way E(B-V) extinction.
        
    """
    def get_query(key, krange=None, merging_logic="or"):
        """ """
        def _build_2d_(kmin,kmax):
            """ """
            if kmin is None and kmax is None:
                return []
            if kmin is None:
                return [f"{key}<={kmax}"]
            elif kmax is None:
                return [f"{key}>={kmin}"]
            else:
                return [f"{kmin}<={key}<={kmax}"]
        
        if krange is None:
            return []
        if np.shape(krange) == (2,):
            return _build_2d_(*krange)
        if np.shape(krange) == (2,2):
            query = [_build_2d_(*krange_) for krange_ in krange]
            return ["("+f" {merging_logic} ".join(np.squeeze(query))+")"]
        
        raise ValueError(f"Cannot for the format of the input krange: {krange}")

    query = []
    # Grid
    gridid = None if (grid is None or grid in ["*","all"]) else get_grid_field(grid)
    query.append([] if gridid is None else ["ID in @gridid"])
    # Ra
    
    query.append(get_query("RA", rarange))
    # Dec
    query.append(get_query("Dec", decrange))
    # gall
    query.append(get_query("GalLong", gallrange))
    # galb
    query.append(get_query("GalLat", galbrange))
    # ecll
    query.append(get_query("EclLong", ecllrange))
    # eclb
    query.append(get_query("EclLat", eclbrange))
    # MW ebmv
    query.append(get_query("Ebv", ebvrange))
    
    
    queries = np.concatenate(query)
    if verbose:
        print(queries)
    if len(queries) == 0:
        return FIELD_DATAFRAME.index
    if verbose:
        print(" & ".join(queries) ) 
    return FIELD_DATAFRAME.query( " & ".join(queries) ).index


def get_field_ccd_qid(ra, dec):
    """ """
    d_ = {}
    field_ccds = get_fields_containing_target(ra,dec, inclccd=True)
    for field_ccd in field_ccds:
        (ramin,decmin),(ramax,decmax) = np.percentile(np.asarray(FIELD_CCDS_GEOSERIE[field_ccd].exterior.xy).T, [0,100], axis=0)
        ccdpos=np.asarray([(ra-ramin)/(ramax-ramin)*6144,(dec-decmin)/(decmax-decmin)*6160])
        qid = ccdpos_to_qid(*ccdpos)
        field, ccd = field_ccd.split("_")
        d_[int(field)] = {"ccd":int(ccd), "qid":int(qid), "rcid":ccdid_qid_to_rcid(int(ccd), int(qid))}
        
    return d_


def spatialjoin_radec_to_fields(radec, fields, how="inner", predicate="intersects",
                                    index_radec="index_radec",**kwargs):
    """ 
    radec: DataFrame or 2d-array 
        coordinates of the points. 
        - DataFrame: must have the "ra" and "dec" columns. 
            This will use the DataFrame's index are data index.
        - 2d array (shape N,2): returned index will be 'range(len(ra))'
    
    fields : [geopandas.geoserie, geopandas.geodataframe or  dict]
        fields contains the fieldid and fields shapes. Several forms are accepted:
        - dict: {fieldid: 2d-array, fieldid: 2d-array ...}
            here, the 2d-array are the field's vertices.

        - geoserie: geopandas.GeoSeries with index as fieldid and geometry as field's vertices.
            
        - geodataframe: geopandas.GeoDataFrame with the 'fieldid' column and geometry as field's vertices.

    Returns
    -------
    GeoDataFrame (geometry.sjoin result )
    """
    import geopandas
    
    # -------- #
    #  Coords  #
    # -------- #
    if type(radec) in [np.ndarray, list, tuple]:
        if (inshape:=np.shape(radec))[-1] != 2:
            raise ValueError(f"shape of radec must be (N, 2), {inshape} given.")
        
        radec = pandas.DataFrame(radec, columns=["ra","dec"])

    # Points to be considered
    geoarray = geopandas.points_from_xy(*radec[["ra","dec"]].values.T)
    geopoints = geopandas.GeoDataFrame({index_radec:radec.index}, geometry=geoarray)
    
    # -------- #
    # Fields   #
    # -------- #
    # goes from dict to geoseries (more natural) 
    if type(fields) is dict:
        values = fields.values()
        indexes = fields.keys()
        # dict of array goes to shapely.Geometry as expected by geopandas
        if type(values.__iter__().__next__()) in [np.ndarray, list, tuple]:
            from shapely import geometry
            values = [geometry.Polygon(v) for v in values]
        
        fields = geopandas.GeoSeries(values,  index = indexes)
            
    if type(fields) is geopandas.geoseries.GeoSeries:
        fields = geopandas.GeoDataFrame({"fieldid":fields.index},
                                        geometry=fields.values)
    elif type(fields) is not geopandas.geodataframe.GeoDataFrame:
        raise ValueError("cannot parse the format of the input 'fields' variable. Should be dict, GeoSeries or GeoPandas")

    # -------- #
    # Joining  #
    # -------- #
    return geopoints.sjoin(fields,  how="inner", predicate="intersects", **kwargs)


def get_fields_containing_target(ra, dec, inclccd=False, buffer=None):
    """ return the list of fields into which the position ra, dec is. 
    Remark that this is based on predefined field positions. 
    Hence, small attrition could affect this.

    Parameters
    ----------
    ra,dec: [float,float]
       coordinates in degree. 
    
    inclccd: [bool] -optional-
        do you want the details of the CCD id on top of the field
        format: "field_ccdid"

    buffer: [float] -optional-
        buffer the polygon (fields or ccds). 
        In degree (unit of the polygons). The inter-ccd gap typically is 0.3 deg

    Returns
    -------
    list (all the field ID that contain the given ra,dec coordinates)
    """
    try:
        from shapely import geometry
    except ImportError:
        raise ImportError("You need shapely to use this function. pip install shapely")
    
    
    coordpoint = geometry.Point(ra, dec)
    fields_geoserie = get_fields_geoserie(inclccd=inclccd)
    if fields_geoserie is None:
        warnings.warn("get_fields_containing_target would be much faster if you install geopandas (pip install geopandas)")
        return [f for f in FIELDSNAMES
                if geometry.Polygon( get_field_vertices(f)[0]).contains(coordpoint)]
    if buffer is None:
        return fields_geoserie.index[ fields_geoserie.contains(coordpoint) ]
    return fields_geoserie.index[ fields_geoserie.buffer(buffer).contains(coordpoint) ]


def get_field_vertices(fieldid=None,
                           inclquad=False, qid=None,
                           inclccd=False, ccd=None,
                           as_dict=False, as_polygon=False,
                        squeeze=True):
    """ Get the fields countours 
    
    Parameters
    ----------
    fieldid: [string or list of] -optional-
        Field (or list of) names as int. 
        If None, all the fields will be used.

    // output format

    as_dict: [bool] -optional-
        Do you want to result as a list (as_dict=False) following the input's fieldid sorting
        or do you want a dict ({fieldid_: verts_ ... })

    as_polygon: [bool] -optional-
        Do you want the vertices as 2d-array (as_polygon=False) or as shapely Geometries (True)

    squeeze: [bool] -optional-
        Should unnecessary dimension be removed ?
        if as_dict is True:
            if squeeze: {fid_ccdid: }
            if not squeeze: {fid:{ccdid: }}
            = if not inclccd, squeez ignored =
        if not as_dict:
           using np.squeeze, basically doing [[]]->[]
    Returns
    -------
    list of dict (see as_dict)
    """
    if fieldid is None:
        fieldid = FIELDSNAMES
        
    if inclccd and ccd is None:
        ccd = np.arange(1,17)
        
    if inclquad and qid is None:
        qid = np.arange(0,64)

    # - Actual calculation
    rafields, decfields  = get_field_centroid( np.asarray(np.atleast_1d(fieldid), dtype="int") ).T
    fields_verts = get_corners(rafields, decfields,
                                inclquad=inclquad, qid=qid,
                                inclccd=inclccd, ccd=ccd,
                                inrad=False, squeeze=False)

    # ----------- #
    #  Format     #
    # ----------- #    
    if as_polygon:
        try:
            from shapely import geometry
            shape = np.shape(fields_verts)
            shape_flat = tuple([np.prod(shape[:-2])] + list(shape[-2:]))
            fields_countours = np.reshape([geometry.Polygon(f_) for f_ in fields_verts.reshape(shape_flat)], shape[:-2])
        except ImportError:
            warnings.warn("You do not have shapely, Please run pip install shapely. 'as_polygon' set to False")
    else:
        fields_countours = fields_verts

    # ----------- #
    #  Output     #
    # ----------- #    
    if not as_dict:
        return fields_countours if not squeeze else np.squeeze(fields_countours)

    # full camera dict
    if inclquad:
        return {f"{field_}_{qid_}": fields_countours[f_][c_]
                    for c_,qid_ in enumerate(np.atleast_1d(qid))
                    for f_,field_ in enumerate(np.atleast_1d(fieldid))}
    
    elif inclccd:
        return {f"{field_}_{ccd_}": fields_countours[f_][c_]
                    for c_,ccd_ in enumerate(np.atleast_1d(ccd))
                    for f_,field_ in enumerate(np.atleast_1d(fieldid))}
    else:
        return {i:k[0] for i,k in zip(fieldid, fields_countours)}
    
def get_field_centroid(fieldid, system="radec"):
    """ Returns the central coordinate [RA,Dec] or  of the given field 

    Parameters
    ----------
    fieldid: [int]
        single field ID

    system: [string] -optional-
        which coordinate system ?
        radec / galactic / ecliptic (default radec)

    Returns
    -------
    [[x_i, y_i],[]]... (depending on your coordinate system)
    Remark if only 1 fieldid given, you have [[x,y]] (not [x,y])
    """
    if system in ["radec", "RADec","RA,Dec", "ra,dec"]:
        syst = ["RA", "Dec"]
    elif system.lower() in ["gal","galactic"]:
        syst = ["Gal Long","Gal Lat"]
    elif system.lower() in ["ecl","ecliptic"]:
        syst = ["Ecl Long","Ecl Lat"]
    else:
        raise ValueError("unknown coordinate system %s select among: [radec / galactic / ecliptic]"%system)
    
    fieldid = np.atleast_1d(fieldid)
    radec = np.asarray(FIELD_DATAFRAME[np.in1d(FIELD_DATAFRAME.index, fieldid)][syst].values)
    
    return radec

def get_corners(ra_field, dec_field, inclquad=False, inclccd=False,
                    qid=None, ccd=None, steps=5, squeeze=True, inrad=False):
    """ """
    from .utils.tools import rot_xz_sph, _DEG2RA

    if inclquad or qid is not None:
        upper_left_corner = _QUAD_COORDS.groupby("Quad").max()
        lower_right_corner = _QUAD_COORDS.groupby("Quad").min()
        if qid is not None:
            upper_left_corner = upper_left_corner.loc[qid]
            lower_right_corner = lower_right_corner.loc[qid]
    # CCD
    elif inclccd or ccd is not None:
        upper_left_corner = _CCD_COORDS.groupby("CCD").max()
        lower_right_corner = _CCD_COORDS.groupby("CCD").min()
        if ccd is not None:
            upper_left_corner = upper_left_corner.loc[ccd]
            lower_right_corner = lower_right_corner.loc[ccd]
    # Focal Plane
    else:
        upper_left_corner = _CCD_COORDS.max()
        lower_right_corner = _CCD_COORDS.min()

        
    ewmin = -np.atleast_1d(upper_left_corner["EW"])
    nsmax = np.atleast_1d(upper_left_corner["NS"])
    ewmax = -np.atleast_1d(lower_right_corner["EW"])
    nsmin = np.atleast_1d(lower_right_corner["NS"])

    ra1  = (np.linspace(ewmax, ewmin, steps)/np.cos(nsmax*_DEG2RA)).T
    dec1 = (np.ones((steps,1))*nsmax).T
    #
    dec2  = np.linspace(nsmax,nsmin, steps).T
    ra2   = ewmin[:,None]/np.cos(dec2*_DEG2RA)
    #
    ra3 = (np.linspace(ewmin,ewmax, steps)/np.cos(nsmin*_DEG2RA)).T
    dec3 = (np.ones((steps,1))*nsmin).T
    #
    dec4  = np.linspace(nsmin,nsmax, steps).T
    ra4 = ewmax[:,None]/np.cos(dec4*_DEG2RA)

    ra_bd = np.concatenate((ra1, ra2, ra3, ra4  ), axis=1)  
    dec_bd = np.concatenate((dec1, dec2, dec3,dec4 ), axis=1)
    
    ra,dec = rot_xz_sph(np.moveaxis(ra_bd,0,1), np.moveaxis(dec_bd,0,1), np.moveaxis(np.atleast_3d(dec_field),0,1))
    ra += np.moveaxis(np.atleast_3d(ra_field),0,1)

    if inrad:
        ra *= _DEG2RA
        dec *= _DEG2RA
        
    radec = np.moveaxis([ra,dec],(0,1,2,3),(3,0,2,1))
    return radec if not squeeze else np.squeeze(radec)

def get_grid_field(which):
    """ """
    if which in ["main","Main","primary"]:
        return FIELDSNAMES[FIELDSNAMES<880]
    if which in ["aux","secondary", "auxiliary"]:
       return FIELDSNAMES[FIELDSNAMES>999]
    if which in ["all","*","both"]:
        return FIELDSNAMES
        
    raise ValueError(f"Cannot parse which field grid you want {which}")


def get_rcid_centroid(rcid, fieldid):
    """ """
    ccdid, q1d = rcid_to_ccdid_qid(rcid)
    return get_qids_centroid(fieldid, ccdid)[f"q{q1d}"]

def get_qids_centroid(fieldid=None, ccdid=None, ccd_vertices=None):
    if ccd_vertices is None:
        if fieldid is None or ccdid is None:
            raise ValueError("either fieldid and ccdid or ccd_vertices should be given")
        
        ccd_vertices = get_field_vertices(fieldid=fieldid, inclccd=True, ccd=ccdid)
        
    min_,mean_,max_ = np.percentile(ccd_vertices, [0,50,100], axis=0)
    return {"q1":np.mean([mean_, max_], axis=0),
            "q2":[np.mean([mean_[0], min_[0]]),np.mean([mean_[1], max_[1]])],
            "q3":np.mean([mean_, min_], axis=0),
            "q4":[np.mean([mean_[0], max_[0]]),np.mean([mean_[1], min_[1]])]
            }
                



def ccdpos_to_qid(ccdx, ccdy):
    """ returns the qid for the given ccd position """
    flagqid = np.asarray(np.asarray([ccdx<3072, ccdy<3080]), dtype=int)    
    return np.asarray([[1,4],[2,3]])[flagqid[0]][flagqid[1]]

def ccdid_qid_to_rcid(ccdid, qid):
    """ computes the rcid """
    return 4*(ccdid - 1) + qid - 1

def rcid_to_ccdid_qid(rcid):
    """ computes the rcid """
    qid = (rcid%4)+1
    ccdid  = int((rcid-(qid - 1))/4 +1)
    return ccdid,qid

##############################
#                            #
#  Fields and References     #
#                            #
##############################
def has_field_reference(fieldid, rcid_details=False, **kwargs):
    """ get the following dictionary {zg:bool, zr:bool, zi:bool}
    where bool is True if the field has a reference image and false otherwise
    
    **kwargs goes to load_metadata(), for instance auth=[username, password]
    Returns
    -------
    {zg:bool, zr:bool, zi:bool}
    """
    from .query import ZTFQuery
    zquery_ = ZTFQuery()
    zquery_.load_metadata(kind="ref", sql_query=f"field={fieldid}", **kwargs)
    if rcid_details:
        return {k:zquery_.metatable.query(f"filtercode in ['{k}']")["rcid"].value_counts(sort=False).to_dict() for k in ["zg", "zr","zi"]}
    return {k: k in zquery_.metatable["filtercode"].values for k in ["zg", "zr","zi"]}

def get_fields_with_band_reference(filter_, ccdid=1, qid=1, **kwargs):
    """ returns the list of fieldid that have a reference image in the `filter_` band.
    filter_ is a filtercode entry [zg, zr or zg]
    
    **kwargs goes to load_metadata(), for instance auth=[username, password]
    Returns
    -------
    list of fieldid
    """
    from .query import ZTFQuery
    zquery_ = ZTFQuery()
    zquery_.load_metadata(kind="ref",
            sql_query="filtercode='%s' and ccdid=%s and qid=%s"%(filter_,ccdid,qid), **kwargs)
    return zquery_.metatable["field"].values

def show_reference_map(band, **kwargs):
    """ Display the 'field plot' in which field with image reference in the given band are colored. """
    title   = "Fields with reference in the %s-band"%band[1]
    field_i = get_fields_with_band_reference(band)
    return show_fields(field_i, facecolor=FIELDNAME_COLOR[band], alpha=0.3, title=title, **kwargs)
    
# ===================== #
#                       #
#    SHOW FIELD         #
#                       #
# ===================== #
def show_fields(fields, vmin=None, vmax=None,
                ax=None, cmap="viridis", title=None,
                colorbar=True, cax=None, hcax=None, clabel=" ", inclhist=True, 
                show_ztf_fields=True, grid="main", grid_prop={},
                bkgd_fields=None, bkgd_prop={},
                show_mw=True, mw_b=None, mw_prop={},
                savefile=None, figsize=None,
                axparam={}, get_fplot=False,
                **kwargs):
    """ 
    Parameters
    ----------
    colored_by: 
    """
    fplot = FieldPlotter(ax=ax, cax=cax, hcax=hcax, figsize=figsize, inclcax=colorbar,
                             inclhist=inclhist, **axparam)
    # - Plotting
    if show_ztf_fields:
        fplot.show_ztf_grid(which=grid, **grid_prop)
        
    if bkgd_fields is not None:
        def_prop = dict(facecolor="0.7", edgecolor="0.5", alpha=0.2, zorder=2)
        fplot.show_fields(bkgd_fields,**{**def_prop,**bkgd_prop})
        
    if show_mw:
        fplot.show_milkyway(b=mw_b, **mw_prop)

    # Removing the NaNs
    fplot.show_fields(fields,
                        colorbar=colorbar,
                        clabel=clabel,cmap=cmap,
                        vmin=vmin, vmax=vmax,**kwargs)
    
    if title is not None:
        fplot.fig.text(0.5,0.9, title,
                     va="top", ha="center", fontsize="large")
    # Output
    if savefile is not None:
        fplot.fig.savefig(savefile, dpi=150)
    if get_fplot:
        return fplot
    return fplot.fig

def show_field_ccds(fieldid, ax=None, ccd=None, textcolor="k", facecolor="0.9", edgecolor="k",
                        autoscale=True, **kwargs):
    """ """
    if ax is None:
        fig = mpl.figure(figsize=[8,6])
        ax = fig.add_subplot(111)
    else:
        fig = ax.figure

    if ccd is None:
        ccd = range(1,17)
    ff = get_fields_geoserie(inclccd=True)
    fccd = {i:ff[f"{fieldid}_{i}"] for i in ccd}
    for i,s_ in fccd.items():
        verts = np.asarray(s_.exterior.xy).T
        ax.add_patch( Polygon(verts, facecolor=facecolor, edgecolor=edgecolor, **kwargs))
        ax.text(*np.mean(verts,axis=0),i, color=textcolor, va="center", ha="center")

    if autoscale:
        ax.autoscale()
        
    return fig

def show_gri_fields(fieldsg=None, fieldsr=None, fieldsi=None,
                    fig=None,
                    title=" ", alignment="horizontal",
                    show_ztf_fields=True, colorbar=True,
                    show_mw=True, mw_b=None, mw_prop={},
                    projection="hammer", moveup=0.05, vscale=1, hscale=1,
                    **kwargs):
    """  """
    prop = {**dict(colorbar=colorbar, edgecolor="0.5", linewidth=0.5),**kwargs}
    
    used_fields = {i+1:f for i,f in enumerate([fieldsg,fieldsr,fieldsi]) if f is not None and len(f)>0}
    
    # None
    if len(used_fields) == 0:
        raise ValueError("No fields given")
    
    # Only one
    if len(used_fields) == 1:
        warnings.warn("Only one color given, favor using show_fields() directly")
        which = list(used_fields.keys())[0]
        return show_fields(used_fields[which], cmap=FIELD_CMAP[which],
                            show_ztf_fields=show_ztf_fields,
                            show_mw=show_mw, mw_b=mw_b, mw_prop=mw_prop,
                            **prop)
    
    # 2 or More
    onlytwo = len(used_fields)==2
    ax, cax = _get_gri_axes_(alignment=alignment, title=title, projection=projection,
                             moveup=moveup, onlytwo=onlytwo, fig=fig, vscale=vscale, hscale=hscale)
    
    fbands = list(used_fields.keys())
    afields_ = list(used_fields.values())
    for i,ax_,cax_,fields_ in zip(fbands, ax, cax, afields_):
        
        if fields_ is not None:
            patch = {}
            if len(np.unique(fields_))==1:
                patch["inclhist"] = False
                if np.unique(fields_)[0]==1:
                    patch["colorbar"] = False
                
            _ = show_fields(fields_, ax=ax_, cax=cax_, cmap=FIELD_CMAP[i],
                            show_ztf_fields=show_ztf_fields,
                            show_mw=show_mw, mw_b=mw_b, mw_prop=mw_prop,
                            **{**prop,**patch})
    return ax[0].figure

def _get_gri_axes_(alignment="classic", title=None, titlefontsize="large", projection="hammer",
                   labelsize="x-small", labelcolor="0.7",  clabelsize="x-small", clabelcolor="k",
                   moveup=None, fig=None, vscale=1, hscale=1, onlytwo=False):
    """ """
    if alignment is None or onlytwo:
        alignment = "flat"

    if moveup is None:
        moveup=0        
        
    if alignment in ["classic"]:
        if fig is None:
            fig = mpl.figure(figsize=[9,6])
        # G
        axg   = fig.add_axes([0.03,0.52+moveup,0.43,0.48], projection=projection)
        caxg  = fig.add_axes([0.03,0.54+moveup,0.43,0.015])
        # R
        axr   = fig.add_axes([0.54,0.52+moveup,0.43,0.48], projection=projection)
        caxr  = fig.add_axes([0.54,0.54+moveup,0.43,0.015])
        # I
        axi   = fig.add_axes([0.27,0.04+moveup,0.43,0.48], projection=projection)
        caxi  = fig.add_axes([0.27,0.05+moveup,0.43,0.015])
        ax = [axg,axr,axi]
        cax = [caxg,caxr,caxi]
    elif alignment in ["flat","aligned", "horizontal"]:
        naxes = 3 if not onlytwo else 2
        if fig is None:
            fig = mpl.figure(figsize=[3.2*naxes,2.5])
        # G
        spanx, spanm = 0.05,0.05
        width = (1-(2*spanx+2*spanm))/naxes
        ax  = [fig.add_axes([ spanm+(i*(width*hscale+spanx)), (0.20+moveup)*vscale, width*hscale, 0.700*vscale],
                                projection=projection) for i in range(naxes)]
        cax = [fig.add_axes([ spanm+(i*(width*hscale+spanx)), (0.12+moveup)*vscale, width*hscale, 0.025*vscale])
                   for i in range(naxes)]
    else:
        raise ValueError(f"cannot parse the given show_gri alignment {alignment}, classic or horizontal")

    if title is not None:
        fig.suptitle(title, fontsize=titlefontsize)
    # labels
    for ax_ in ax:
        ax_.tick_params(labelsize=labelsize, labelcolor=labelcolor)
    for ax_ in cax:        
        ax_.tick_params(labelsize=clabelsize, labelcolor=clabelcolor)
        
    return ax,cax


def show_ztf_fieldvalues(key="Ebv", fieldid="main", mindec=-30,
                        vmin=None, vmax=None,
                        ax=None, cmap="viridis", title=None,
                        colorbar=True, cax=None, clabel=" ",
                        show_ztf_fields=True, grid="main", grid_prop={},
                        show_mw=True, mw_b=None, mw_prop={},
                        savefile=None, **kwargs):
    """ """
    if key not in FIELD_DATAFRAME.columns:
        raise ValueError(f"cannot parse the given key {key}, only columns from FIELD_DATAFRAME available")
    
    
    if type(fieldid) is str or fieldid is None:
        fieldid = get_grid_field(fieldid)
    query_ = "index in @fieldid"
    if mindec is not None:
        query_ +=" and Dec > @mindec"
    # Serie to plot 
    serie = FIELD_DATAFRAME.query(query_)[key]
    return show_fields(fields=serie, vmin=vmin, vmax=vmax,
                           ax=ax, cmap=cmap, title=title,
                           colorbar=colorbar, cax=cax, clabel=clabel,
                           show_ztf_fields=show_ztf_fields, grid=grid, grid_prop=grid_prop,
                           show_mw=show_mw, mw_b=mw_b, mw_prop=mw_prop,
                           savefile=savefile, **kwargs)
    
    
def display_field(ax, fieldid, origin=None, facecolor="0.8", lower_dec=None, edgecolor=None, **kwargs):
    """ """
    print("display_field is DEPRECATED")

def _radec_to_plot_(self, ra, dec):
    """ """
    return np.asarray([-(np.asarray(ra)-self.origin)*np.pi/180, np.asarray(dec)*np.pi/180])

        
##############################
#                            #
#  Individual Field Class    #
#                            #
##############################
class FieldPlotter( object ):
    """ """
    def __init__(self, ax=None, origin=180, inclcax=True, inclhist=False, cax=None, hcax=None, **kwargs):
        """ """
        self.origin = origin        
        self.load_ax(ax, inclcax=inclcax, cax=cax, inclhist=inclhist, **kwargs)

        
    def load_ax(self, ax=None, update_ticks=True, cax=None, hcax=None, inclcax=True, inclhist=False, figsize=None,
                    **kwargs):
        """ """
        if ax is None or len(np.atleast_1d(ax))==4:
            self.fig = mpl.figure(figsize=(8,5) if figsize is None else figsize)
            self.ax = self.fig.add_axes([0.15,0.2,0.75,0.75] if ax is None else ax, projection="hammer")
        else:
            self.ax = ax
            self.fig = self.ax
            
        if inclcax:
            from .utils.plots import HistColorbar
            if cax is None and hcax is None:
                from .utils.plots import insert_ax
                cax = insert_ax(self.ax, "bottom",
                                    shrunk=0.98, space=-0.15, axspace=0.13)
            if inclhist and hcax is None:
                if len(np.atleast_1d(cax)) == 4:
                    xmin, ymin, width, height= cax
                    hcax = [xmin, ymin+np.min([height*1.5,height+0.005]), width, height*1.8]
                else:    
                    bcax = cax.get_position()
                    xmin, ymin, width, height= bcax.xmin, bcax.ymin, bcax.width, bcax.height
                    hcax = cax.figure.add_axes([xmin, ymin+np.min([height*1.5,height+0.005]) , width, height*1.8])
                    
            self.histcbar = HistColorbar(ax=hcax, cax=cax, fig=self.fig, draw=False)
        else:
            self.histcbar = None
            
        if update_ticks:
            tick_labels = np.array([150, 120, 90, 60, 30, 0, 330, 300, 270, 240, 210])
            tick_labels = np.remainder(tick_labels+360+self.origin,360)
            self.ax.set_xticklabels(tick_labels)     # we add the scale on the x axis

        # Label param
        if len(kwargs)>0:
            self.ax.tick_params(**kwargs)
            
    # ---------- #
    #  PLOTTER   #
    # ---------- #
    def show_ztf_grid(self, which="main", 
                  facecolor="None", edgecolor="0.7", alpha=0.1, zorder=1, **kwargs):
        """ """
        fields_ = get_grid_field(which)
        self._ztfgrid = self.show_fields(fields_, 
                                             facecolor=facecolor, edgecolor=edgecolor, 
                                             alpha=alpha, zorder=zorder, **kwargs)

    def show_milkyway(self, b=None, nbins=100, l_start=-241, l_stop=116,**kwargs):
        """ """
        
        nbins=100
        if b is None:
            gal = coordinates.Galactic(np.linspace(l_start,l_stop,nbins)*units.deg, np.zeros(nbins)*units.deg
                                           ).transform_to(coordinates.ICRS)
            prop = dict(ls="-", color="0.7", alpha=0.5)
            self.ax.plot(*self.radec_to_plot(gal.ra, gal.dec), **{**prop, **kwargs})
        else:
            gal_dw = coordinates.Galactic(np.linspace(l_start,l_stop,100)*units.deg, +b*np.ones(nbins)*units.deg
                                         ).transform_to(coordinates.ICRS)
            gal_up = coordinates.Galactic(np.linspace(l_start,l_stop,100)*units.deg, -b*np.ones(nbins)*units.deg
                                         ).transform_to(coordinates.ICRS)
            ra_dw,dec_dw = self.radec_to_plot(gal_dw.ra, gal_dw.dec)
            ra_up,dec_up = self.radec_to_plot(gal_up.ra, gal_up.dec)

            prop = dict(facecolor="0.7", alpha=0.2)
            self.ax.fill_between(ra_dw, dec_dw, dec_up, **{**prop, **kwargs})
        
    def add_fields(self, fields, facecolor="0.7", edgecolor="k", lw=0.5, **kwargs):
        """ """
        fields_verts = self.get_field_vertices(fields)
        self.poly_ = [self.ax.add_patch(Polygon(p_, facecolor=facecolor, 
                                        edgecolor=edgecolor, lw=lw, **kwargs))
                     for p_ in fields_verts]

    def show_fields(self, fields, cmap=None,
                        colorbar=True, clabel=None, cfontsize=None,
                        vmin=None, vmax=None, bins="auto", **kwargs):
        """ fields could be a list of field or a dictionary with single values """
        
        # For now, then will move to pandas.Series as native
        if type(fields) is pandas.Series:
            fields = fields.to_dict()
            
        if type(fields)==dict:
            # popint out nans
            fields = {f:v for f,v in fields.items() if not np.isnan(v)} 
            values = list(fields.values())
            if len(values)==0 or not np.any(values):
                if self.histcbar is not None:
                    self.histcbar.set_visible(False)
                return
            
            if vmin is None: vmin = "0"
            if type(vmin) == str:
                vmin=np.percentile(values, float(vmin))
            if vmax is None: vmax = "100"
            if type(vmax) == str:
                vmax=np.percentile(values, float(vmax))

            values = np.asarray(values)
            if self.histcbar is not None:
                if bins is None or bins in ['auto']:
                    if values.dtype == int:
                        bins = int((vmax-vmin))+1
                        vmax +=1
                self.histcbar.build_histrogram(data=values, vmin=vmin, vmax=vmax, bins=bins)
                
            # - CBAR
            if cmap is None and self.histcbar is not None:
                cmap = self.histcbar.cmap
                
            elif type(cmap) == str:
                if self.histcbar is not None:
                    self.histcbar.load_cmap(cmap)
                    cmap = self.histcbar.cmap
                else:
                    cmap = mpl.get_cmap(cmap)
            else:
                self.histcbar.load_cmap(cmap.name)
                
                
#            _ = kwargs.pop("facecolor",None) #remove facecolor is any
            for f,v in fields.items():
                self.add_fields(f, **{**dict(facecolor=cmap((v-vmin)/(vmax-vmin)) if vmax-vmin !=0 else cmap(0)), **kwargs})

            # - Cbar
            if colorbar:
                self.show_colorbar(clabel=clabel, cfontsize=cfontsize)
            elif self.histcbar is not None:
                 self.histcbar.set_visible(False)
                 
        else:
            self.add_fields(fields, **kwargs)
        
    def show_point(self, radec, **kwargs):
        """ """
        xy = self.radec_to_plot(*radec)
        self.ax.scatter(*xy, **kwargs)

    def show_colorbar(self, clabel=None, cfontsize=None):
        """ """
        if self.histcbar is None:
            return

        self.histcbar.draw()
        self.histcbar.set_label(clabel, fontsize=cfontsize)
        
    def get_field_vertices(self, fields_):
        """ Get the field vertices in plotting coordinates. """
        fields_ = np.squeeze(get_field_vertices(fields_, squeeze=False), axis=1) # remove CCDs
        return [self.radec_to_plot(*f_.T).T for f_ in fields_] 
    
    def radec_to_plot(self, ra, dec):
        """ """
        return np.asarray([-(np.asarray(ra)-self.origin)*np.pi/180, np.asarray(dec)*np.pi/180])


   
##############################
#                            #
#    ZTF Fields Class        #
#                            #
##############################

class FieldAnimation( FieldPlotter ):
    
    def __init__(self, fields, ax=None,dates=None, facecolors=None, alphas=None, edgecolors=None, inclcax=False):
        """ """
        super().__init__(ax=ax, inclcax=inclcax)
        
        self.set_fields(fields)
        self.set_dates(dates)
        self.set_properties(facecolors=facecolors, alphas=alphas, edgecolors=edgecolors)
        
    # ================= #
    #   Methods         #
    # ================= #
    def set_dates(self, dates):
        """ """
        self._dates = np.atleast_1d(dates) if dates is not None else None

    def set_fields(self, fields):
        """ """
        self._fields = fields
        self._unique_fields = np.unique(self._fields)
        self._field_vertices = {i:v for i,v in zip(self._unique_fields,self.get_field_vertices(self._unique_fields))}
        
    def set_properties(self, facecolors=None, edgecolors=None, alphas=None):
        """ """
        self._display_prop = {}
        self._set_prop_("facecolor", facecolors, "0.7")
        self._set_prop_("edgecolor", edgecolors, "None")
        self._set_prop_("alpha", alphas, 1)

    def _set_prop_(self, key, value, default=None):
        """ """
        if not hasattr(self,"_display_prop"):
            self._display_prop = {}
            
        if value is None:
            value = default
            
        if len(np.atleast_1d(value)) == 1:
            self.display_prop[key] = np.atleast_1d(value)
            self.display_prop[f"unique_{key}"] = True
        else:
            self.display_prop[key] = value
            self.display_prop[f"unique_{key}"] = False
            
    # ---------- #
    #  SETUP     #
    # ---------- #
    def reset(self):
        """ """
        self.intpoly_ = Polygon(self.field_vertices[self.fields[0]],
                                    facecolor=self.display_prop["facecolor"][0],
                                    edgecolor=self.display_prop["edgecolor"][0],
                                    alpha=self.display_prop["alpha"][0])
        
        if self._dates is not None:
            self.inttext_ = self.fig.text(0.01,0.99, self._dates[0],
                                           va="top", ha="left", weight="bold")
            
        p_ = self.ax.add_patch(self.intpoly_)
        return self.intpoly_
        

    # ---------- #
    #  Animate   #
    # ---------- #    
    def update_field_to(self, i):
        """ """
        try:
            self.intpoly_.set_xy(self.field_vertices[ self.fields[i] ])
            if self.dates is not None and len(self.dates)>1:
                self.inttext_.set_text(self.dates[i])
        except:
            print(f"FAILES for i={i}")
            
        for key in ["facecolor","edgecolor","alpha"]:
            if not self.display_prop[f"unique_{key}"]:
                getattr(self.intpoly_,f"set_{key}")(self.display_prop[key][i])
                
        return self.intpoly_

    def launch(self, interval=5, repeat=False, blit=True, savefile=None):
        """ """
        from matplotlib import animation
        self.anim = animation.FuncAnimation(self.fig, self.update_field_to,
                                                init_func=self.reset,
                               frames=self.nfields, interval=interval, repeat=repeat, blit=blit)
        
    # ================= #
    #   Properties      #
    # ================= #
    @property
    def fields(self):
        """ Fields that should be shown """
        return self._fields

    @property
    def dates(self):
        """ Observation dates if any """
        return self._dates
    @property
    def nfields(self):
        """ size of self.fields """
        return len(self.fields)

    @property
    def field_vertices(self):
        """ vertices of the fields """
        return self._field_vertices

    @property
    def display_prop(self):
        """ """
        return self._display_prop


    
##############################
#                            #
#    Planner Class           #
#                            #
##############################

class PalomarPlanning( object):
    """ """
    def __init__(self, date=None, **kwargs):
        """ """
        self._site = coordinates.EarthLocation.of_site("palomar")
        self._utcshift = -8*units.h
        if date is not None:
            self.set_date(date, **kwargs)

    # --------- #
    #  SETTER   #
    # --------- #
    def set_date(self, date, timerange=[-7,7], to_utc=True, **kwargs):
        """ """
        self._date = date
        self.set_night(date,timerange=[-7,7], to_utc=True, **kwargs)
        
    def set_night(self, night, timerange=[-7,7], to_utc=True, **kwargs):
        """ """    
        self._night = self.get_night(night, timerange=timerange, to_utc=to_utc, **kwargs)
        self._nightaltaz = self._get_night_altaz_(self.night)
        
    # --------- #
    #  GETTER   #
    # --------- #
    @classmethod
    def get_date_night_duration(cls, date, twilight=-12*units.deg, **kwargs):
        """ ClassMethod to directly get the night duration at a given date or list of dates. """
        return cls().get_night_duration(date, twilight=twilight, **kwargs)

    
    def get_night(self, date, timerange=[-7,7], to_utc=True, range_units=units.h, bins=100, **kwargs):
        """ """
        if type(date) is not time.Time:
            date = time.Time(date, **kwargs)
        if to_utc:
            date -= self.utcshift
            
        if timerange is None:
            return date
        if len(np.atleast_1d(date))==1:
            return date + np.linspace(*timerange, bins)*range_units
        return date + np.linspace(*timerange, bins)[:,None]*range_units

    def get_night_duration(self, date=None, twilight=-12*units.deg, bins=500):
        """ in hours """
        if date is None:
            date =self.date
        
        time_, sunaltaz = self.get_body_altaz("sun",  date, bins=bins)
        flagnight = sunaltaz.alt<twilight
        
        if len(np.atleast_1d(date))==1:
            nigh_time = time_[flagnight][[0,-1]].jd
            return np.diff(nigh_time)[0]*units.day.to("h")*units.h
        
        length = []
        for i,d_ in enumerate(date):
            nigh_time_ = time_[:,i][flagnight[:,i]][[0,-1]].jd
            night_length = np.diff(nigh_time_)[0]*units.day.to("h")*units.h
            length.append(night_length)
        return length

    def get_fields_altaz(self, fieldid, date=None, **kwargs):
        """ """
        radec = get_field_centroid(fieldid)
        return self.get_coord_altaz(radec, date=date, **kwargs)

    def get_coord_altaz(self, radec, date=None, **kwargs):
        """ """
        if type(radec) is not coordinates.SkyCoord:
            radec = coordinates.SkyCoord(radec, unit="deg")
            
        date, datealtaz = self._read_date_input_(date, **kwargs)
            
        return date, radec.transform_to( datealtaz[:,None] )
    
    def get_body_altaz(self, bodyname, date=None, **kwargs):
        """ 
        'sun', 'moon', 'mercury', 'venus', 'earth-moon-barycenter', 'mars', 'jupiter', 'saturn', 'uranus', 'neptune'
        """
        date, datealtaz = self._read_date_input_(date, **kwargs)
        body = coordinates.get_body(bodyname, date)
        return date, body.transform_to( datealtaz )
        
        
    def get_fields_observability(self, fieldid, date=None, airmasslimit=[1,1.5], minobservability=90*units.min,
                                     twilight=-18*units.deg, still_observablein=None):#2*units.week):
        """ """
        nighttime, field_altaz = self.get_fields_altaz(fieldid, date=date)

        airmasses = np.asarray(field_altaz.secz)
        # Twilight Cut
        flagtwilight = self.is_twilight(date, twilight, squeeze=True)
        flag_good = ((~flagtwilight[:,None]) * (airmasses>=airmasslimit[0]) * (airmasses<airmasslimit[1]) )
        if minobservability is not None:
            min_night_frac = minobservability/(nighttime[~flagtwilight].max()-nighttime[~flagtwilight].min()).to("min")
            flag_nightfrac = (np.sum(flag_good, axis=0)/len(nighttime[~flagtwilight])>=min_night_frac)
            flag_good *= flag_nightfrac[None,:]


        stime, ffrac = [pandas.Series(np.sum(flag_good, axis=1), index=pandas.DatetimeIndex(nighttime.datetime)),
                        pandas.Series(np.sum(flag_good, axis=0)/len(nighttime[~flagtwilight]), index=fieldid)]
        
        if still_observablein is not None:
            print("using still_observablein options") 
            if date is None:
                date = self.date
                
            delay_date = (time.Time(date) + still_observablein).iso.split(" ")[0]
            delay_stime, delay_ffrac = self.get_fields_observability(fieldid, date=delay_date,
                                                                    airmasslimit=airmasslimit, minobservability=minobservability,
                                                                    twilight=twilight, still_observablein=None)
            fieldidstillin = ffrac[ffrac>0].index[np.in1d(ffrac[ffrac>0].index,delay_ffrac[delay_ffrac>0].index)]
            flagstillin = np.in1d(ffrac.index,fieldidstillin)
            flag_good *= flagstillin[None,:]
            
        return [pandas.Series(np.sum(flag_good, axis=1), index=pandas.DatetimeIndex(nighttime.datetime)),
                pandas.Series(np.sum(flag_good, axis=0)/len(nighttime[~flagtwilight]), index=fieldid)]
    
        
    def get_observable_fields(self, fieldids, date=None, airmasslimit=[1,1.5],
                                minobservability=90*units.min, twilight=-18*units.deg,
                                **kwargs):
        """ """
        stime, ffrac = self.get_fields_observability(fieldids, date=date, airmasslimit=airmasslimit,
                                                     twilight=twilight, minobservability=minobservability,
                                                     **kwargs)
        return ffrac[ffrac>0].index
    
        
    def is_twilight(self, date=None, sunlimit=-18*units.deg, squeeze=True):
        """ """
        time_, sunaltaz = self.get_body_altaz("sun", date)
        sunlimit = np.atleast_1d(sunlimit)
        
        flags = [sunaltaz.alt>slimit for slimit in sunlimit]
        
        if len(sunlimit) ==1 and squeeze:
            return flags[0]
        return flags
    
    def is_day(self, date, sunlimit=0*units.deg, squeeze=True, **kwargs):
        """ """
        return self.is_twilight(sunlimit=0*units.deg, squeeze=squeeze, **kwargs)

    def _get_night_altaz_(self, night, **kwargs):
        """ 
        **kwargs goes to astropy.time.Time if date is not one already.
        """
        return coordinates.AltAz(obstime=night, location=self.site)
    
    def _read_date_input_(self, date, **kwargs):
        """ """
        if date is not None:
            if type(date) is not time.Time:
                date = self.get_night(date, **kwargs)
            datealtaz = self._get_night_altaz_(date)
        elif self.has_night():
            date      = self.night
            datealtaz = self.nightaltaz
        else:
            raise ValueError("No night set (self.set_night), no date given (see date option).")
            
        return date, datealtaz
    # -------- #
    #  PLOTTER #
    # -------- #
    def show(self, date=None, fields=None, radec=None, body=None, propdate={},
            show_twilight=True, as_airmass=False, ctwilight="0.7"):
        """ """
        import matplotlib.pyplot as mpl
        from matplotlib import dates as mdates
        
        fig = mpl.figure(figsize=[7,4])
        ax = fig.add_subplot(111)

        if date is not None:
            if type(date) is not time.Time:
                date = self.get_night(date, **propdate)
        else:
            date = self.night
            
        # 
        # - start: Plotting
        for type_ in ["fields", "radec", "body"]:
            t_ = eval(type_)
            if t_ is not None:
                _, v_ = getattr(self,f"get_{type_}_altaz")(t_, date)
                ax.plot(date.datetime, v_.alt if not as_airmass else v_.secz)
        # - end: Plotting
        # 
            
        # 
        # - start: Twilight bands
        if show_twilight:
            self.show_twilight(ax=ax, date=date)
            
        # - start: End bands
        # 
        
        # - 
        ax.set_ylabel("Altitude [deg]" if not as_airmass else "Airmass")
        # - cleaning
        locator = mdates.AutoDateLocator()
        formatter = mdates.ConciseDateFormatter(locator)
        ax.xaxis.set_major_locator(locator)
        ax.xaxis.set_major_formatter(formatter)
        # - Limites
        ax.set_xlim(date.datetime[0],date.datetime[-1])
        
        ax.set_ylim(-1,3) if as_airmass else ax.set_ylim(-10,95)
        
    def show_twilight(self, date=None, ax=None, sunlimit=[0,-12,-18]*units.deg,
                     ctwilights="0.7", alphas=[0.8,0.3,0.1], propdate={}):
        """ """
        import matplotlib.pyplot as mpl
                
        ctwilights = np.atleast_1d(ctwilights)
        alphas = np.atleast_1d(alphas)
        
        if len(ctwilights)==1:
            ctwilights = list(ctwilights)*len(sunlimit)
        if len(alphas)==1:
            alphas = list(alphas)*len(alphas)
            
        if ax is None:
            fig = mpl.figure(figsize=[7,4])
            ax = fig.add_subplot(111)
        else:
            fig = ax.figure
            
        
        if date is not None:
            if type(date) is not time.Time:
                date = self.get_night(date, **propdate)
        else:
            date = self.night
        
        for i, tflag in enumerate(self.is_twilight(date=date, sunlimit=sunlimit, squeeze=False)):
            ax.fill_between(date.datetime, 0,1, where=tflag, 
                            transform=ax.get_xaxis_transform(), 
                            color=ctwilights[i], alpha=alphas[i])
        
        
    def show_fields_observability(self, fieldids, show_twilight=True, cmap="viridis",
                                      airmasslimit=[1,1.5], twilight=-18*units.deg, minobservability=90*units.min,
                                      show_notobs=True, cmapv=0.1,
                                      fcno="0.7",ecno="0.7", alphano=0.15, nolw=0, noprop={},
                                      **kwargs):
        """ 
        **kwargs foes to get_fields_observability() 
        """
        import matplotlib.pyplot as mpl        
        from matplotlib import dates as mdates

        figsize = [9,3.5]
        axmap   = [0.05,0.21,0.45,0.7]
        caxmap  = [0.05,0.18,0.45,0.02]

        stime, ffrac = self.get_fields_observability(fieldids, airmasslimit=airmasslimit, twilight=twilight, minobservability=minobservability)
        
        fig = show_fields( ffrac[ffrac.values>0],
                               bkgd_fields=fieldids[ffrac.values==0] if show_notobs else None,
                               bkgd_prop={**dict(facecolor=fcno, edgecolor=ecno, alpha=alphano, lw=nolw),**noprop},
                                  ax=axmap, cax=caxmap, figsize=figsize, cmap=cmap,
                                  axparam={"labelsize":"x-small","color":"0.7", "labelcolor":"0.7"},
                                  clabel="Fraction of night observable", cfontsize="medium")

        axh = fig.add_axes([0.6,0.17,0.35,0.68])
        axh.fill_between(stime.index, stime.values, 
                         facecolor=mpl.cm.get_cmap(cmap)(cmapv,0.2), 
                         edgecolor=mpl.cm.get_cmap(cmap)(cmapv,0.9), lw=2, **kwargs)


        locator = mdates.AutoDateLocator(minticks=4, maxticks=6)
        formatter = mdates.ConciseDateFormatter(locator)
        axh.xaxis.set_major_locator(locator)
        axh.xaxis.set_major_formatter(formatter)
        axh.set_ylim(bottom=0)
        axh.set_xlim(*self.night.datetime[[0,-1]])
        axh.set_ylabel("Number of fields observable")
        if show_twilight:
            self.show_twilight(ax=axh, date=self.night, ctwilights="0.9")
            
        fig.text(0.02,0.98, f"Field Observability | {len(ffrac[ffrac>0])} fields", color="0.7", fontsize="medium", va="top",ha="left")
        
        return fig
    
    # ============= #
    #  Properties   #
    # ============= #
    @property
    def site(self):
        """ """
        return self._site
    @property
    def utcshift(self):
        """ """
        return self._utcshift

    @property
    def date(self):
        """ """
        return None if not hasattr(self,"_date") else self._date
    
    @property
    def night(self):
        """ """
        if not self.has_night():
            return None
        return self._night
    
    def has_night(self):
        """ """
        return getattr(self,"_night") and self._night is not None
    
    @property
    def nightaltaz(self):
        """ """
        if not self.has_night():
            raise AttributeError("No night set.")
            
        return self._nightaltaz



#! /usr/bin/env python
#

"""  Access SEDM data from pharos """


from . import io
import warnings
import pandas
import numpy as np
import json
import requests
import os
PHAROS_BASEURL = "http://minar.caltech.edu"
E3D_BASE = 'e3d_crr_b_'
GUIDER_BASE = 'guider_crr_b_'
SPEC_BASE_AUTO = "spec_auto_robot_lstep1__crr_b_"
SPEC_BASE_CONTSEP = "spec_auto_contsep_lstep1__crr_b_"

SEDMLOCAL_BASESOURCE = os.path.join(io.LOCALSOURCE, "sedm")
SEDMLOCALSOURCE = os.path.join(SEDMLOCAL_BASESOURCE, "redux")
if not os.path.exists(SEDMLOCAL_BASESOURCE):
    os.makedirs(SEDMLOCAL_BASESOURCE, exist_ok=True)
if not os.path.exists(SEDMLOCALSOURCE):
    os.makedirs(SEDMLOCALSOURCE, exist_ok=True)


#######################
#                     #
#  High level method  #
#                     #
#######################

#
# - whatfiles
#
def download_whatfile(date, as_dataframe=True, auth=None, store=True):
    """ Download summary of files observed on the given date

    Parameters
    ----------
    date: [string]
        Date with the following format: YYYYMMDD

    as_dataframe: [bool] -optional-
        output format: text (False) or DataFrame (True)
    """
    whatfile = _download_sedm_data_(
        date, "what.list", auth=auth).text.splitlines()

    if as_dataframe or store:
        data = whatfiles_to_dataframe(whatfile)
        if store:
            fileout = get_whatfile_path(date)
            if not os.path.isdir(os.path.dirname(fileout)):
                os.makedirs(os.path.dirname(fileout), exist_ok=True)

            data.to_parquet(fileout)

    return whatfile if not as_dataframe else data


def download_pharosfile(date, auth=None, store=True):
    """ """

    pharosfile = _download_sedm_pharosfile_(date, auth=auth)

    if store:
        fileout = get_pharosfile_path(date)
        if not os.path.isdir(os.path.dirname(fileout)):
            os.makedirs(os.path.dirname(fileout), exist_ok=True)

        with open(fileout, "w") as fileout_:
            for l_ in pharosfile:
                fileout_.write(l_+"\n")

    return pharosfile


def bulk_download(which, dates, client=None):
    """ """
    if which not in ["pharosfile", "whatfile"]:
        raise ValueError(
            f"which can either be 'pharosfile' or 'whatfile', {which} given")

    download_func = eval(f"download_{which}")
    if client:
        from dask import delayed
        d_download = [delayed(download_func)(date_) for date_ in dates]
        return client.compute(d_download)

    return [download_func(date_) for date_ in dates]


def get_whatfile_path(date, extension=None):
    """ """
    if extension is None:
        extension = ".parquet"
    if not extension.startswith("."):
        extension = f".{extension}"

    if date is None or date == "stored":
        return os.path.join(SEDMLOCAL_BASESOURCE, "whatfiles", f"stored_data{extension}")
    return os.path.join(SEDMLOCAL_BASESOURCE, "whatfiles", f"what_{date.lower()}{extension}")


def get_pharosfile_path(date, extension=None):
    """ """
    if extension is None:
        extension = ".dat"

    if not extension.startswith("."):
        extension = f".{extension}"

    return os.path.join(SEDMLOCAL_BASESOURCE, "pharosfiles", f"pharosfile_{date.lower()}{extension}")

# Internal Low level download


def _download_sedm_data_(night, pharosfile, fileout=None, auth=None, verbose=False):
    """ """
    url = os.path.join(PHAROS_BASEURL, "data", night, pharosfile)
    if verbose:
        print(url)
    if auth is None:
        auth = io._load_id_("pharos")
    return io.download_single_url(url, fileout=fileout,
                                  auth=auth,
                                  cookies="no_cookies")


def _download_sedm_pharosfile_(date, auth=None):
    """ """
    username, password = io._load_id_("pharos") if auth is None else auth
    requests_prop = {"data": json.dumps({"obsdate": date,
                                         "username": username,
                                         "password": password,
                                         }),
                     "headers": {'content-type': 'application/json'}}

    t = requests.post(os.path.join(
        PHAROS_BASEURL, "get_user_observations"), **requests_prop).text
    try:
        t = json.loads(t)
    except:
        raise IOError("Cannot load the downloaded file (not a json) ")

    if "data" not in t:
        if "message" in t:
            if ("Could not find summary file" in t["message"]) or ("No data directory could b" in t["message"]):
                return []

        raise IOError(
            "night file download fails. Check you authentification maybe?")

    return np.sort(t["data"])


############################
#                          #
# Download from Whatdatas  #
#                          #
############################

def get_pharos_urls_from_whatdata(df, kind, contains=None):
    """ Get Pharos URL's from whatdatas dataframe. 

        Parameters
        ----------
        df: [DataFrame]
            Whatdatas dataframe 

        kind: [str]
            Might be e3d/cube or astrom/guider or HexaGrid/HexaGrid.pkl

        contains: [str] -optional-
            Only select the filenames where *contains* appears.

        Returns
        -------
        Pharos URLS """

    urls = []
    if contains is not None:
        df = df[np.logical_or(df['filename'].str.contains(
            contains), df['target'].str.contains(contains))]

    for f in range(len(df)):

        if kind in ['e3d', 'cube']:
            urls.append(os.path.join(PHAROS_BASEURL, 'data',
                        df.iloc[f].name[0],  E3D_BASE + df.iloc[f]['filename'].rsplit('.')[0]+'_'+df.iloc[f]['target']+'.fits'))

        if kind in ['HexaGrid', 'HexaGrid.pkl']:
            urls.append(os.path.join(PHAROS_BASEURL, 'data',
                        df.iloc[f].name[0], df.iloc[f].name[0]+'_HexaGrid.pkl'))

        if kind in ['spec', 'spectra']:
            urls.append(os.path.join(PHAROS_BASEURL, 'data',
                        df.iloc[f].name[0],  SPEC_BASE_AUTO + df.iloc[f]['filename'].rsplit('.')[0]+'_'+df.iloc[f]['target']+'.fits'))
            urls.append(os.path.join(PHAROS_BASEURL, 'data',
                        df.iloc[f].name[0],  SPEC_BASE_CONTSEP + df.iloc[f]['filename'].rsplit('.')[0]+'_'+df.iloc[f]['target']+'.fits'))

        if kind in ['astrom', 'guider']:
            urls.append(os.path.join(PHAROS_BASEURL, 'data',
                        df.iloc[f].name[0],  GUIDER_BASE + df.iloc[f]['filename'].rsplit('.')[0]+'_' + 'astrom.fits.gz'))
            urls.append(os.path.join(PHAROS_BASEURL, 'data',
                        df.iloc[f].name[0],  GUIDER_BASE + df.iloc[f]['filename'].rsplit('.')[0]+'_' + 'astrom.fits'))
    return urls


def get_local_path_from_pharos_urls(urls):
    """ Get SEDMLocalSource path from pharos urls. 

        Parameters
        ----------
        urls: [str]
            Pharos URL(S) to convert.

        Returns
        -------
        SEDMLocalSource path (list of str) """
    from pysedm.io import parse_filename
    urls = np.atleast_1d(urls)
    local_url = []
    for url in urls:

        if 'HexaGrid' in url:
            date = url.rsplit('/')[-1].rsplit('_')[0]
            local_url.append(os.path.join(
                SEDMLOCALSOURCE, date,  os.path.basename(url)))

        elif 'spec' in url:
            date = url.rsplit('/')[-2]
            local_url.append(os.path.join(
                SEDMLOCALSOURCE, date,  os.path.basename(url)))

        else:
            info = parse_filename(os.path.basename(url))

            if os.path.basename(url).rsplit('ifu')[0].startswith('e3d'):
                base = E3D_BASE
                extension = info['name'] + '.fits'
            elif os.path.basename(url).rsplit('ifu')[0].startswith('guider'):
                base = GUIDER_BASE
                if os.path.basename(url).endswith('.fits'):
                    extension = info['name'] + '.fits'
                elif os.path.basename(url).endswith('.fits.gz'):
                    extension = info['name'] + '.fits.gz'
            else:
                raise OSError(
                    f'Only guider files and e3d cubes files are implemented. URL basename should starts with "guider" or "e3d", but starts with{os.path.basename(url)[:10]} ')

            local_url.append(os.path.join(
                SEDMLOCALSOURCE, info['date'], base + info['sedmid'] + '_' + extension))
    return local_url


def _download_sedm_data_from_url(url, fileout=None, return_filename=True, check_file=True, auth=None, verbose=False, overwrite=False, show_progress=False):
    """ Download sedm data from single url. 

        Parameters
        ----------
        url: [str]
            Pharos URL from which you want the data.

        fileout: [str] -optional-
            fileout path to save the datas. If None (Default) or 'local', save the datas at the SEDMLOCALSOURCE.

        Returns
        -------
        Fileout if return_filename is True, else None """

    if verbose:
        print(url)
    if auth is None:
        auth = io._load_id_("pharos")
    if fileout in [None, 'local']:
        fileout = get_local_path_from_pharos_urls(url)[0]

    io.download_single_url(url, fileout=fileout,
                           auth=auth, overwrite=overwrite, show_progress=show_progress,
                           cookies="no_cookies")
    if check_file:
        from astropy.io import fits
        if '.fits' in fileout:
            try:
                _ = fits.getdata(fileout)
            except FileNotFoundError:
                warnings.warn(f"[Errno 2] No such file or directory: {fileout}")
                fileout = None
    if return_filename:
        return fileout


def download_from_whatdata(df, kind, contains=None, client=None, dirout=None, auth=None, verbose=False, overwrite=False, show_progress=False, return_filename=True, check_file=True,):
    """ Main Download sedm data method from whatdata DataFrame. 

        Parameters
        ----------
        df: [DataFrame]
            Whatdatas dataframe 

        kind: [str]
            Might be e3d/cube or astrom/guider

        contains: [str] -optional-
            Only select the filenames where *contains* appears.

        client: [dask.distributed.client] -optional-
            Client to use if you want to download with Dask.

        dirout: [str] -optional-
            Directory to save the datas. If None (Default), save the datas at the SEDMLOCALSOURCE.

        Returns
        -------
        Fileouts if return_filename is True, else None """

    pharos_urls = get_pharos_urls_from_whatdata(
        df=df, kind=kind, contains=contains)
    if dirout is not None:
        fileouts = os.path.joint(dirout, os.path.basename(pharos_urls))
    else:
        fileouts = np.tile(None, len(pharos_urls))
    filenames = []
    if client is not None:
        from dask import delayed, distributed
        for url, fileout in zip(pharos_urls, fileouts):
            filenames.append(delayed(_download_sedm_data_from_url)(
                url, fileout, return_filename, check_file, auth, verbose, overwrite, show_progress))
        res = client.compute(filenames)
        distributed.wait(res)
        return [res[f].result() for f in range(len(res))] if return_filename else None
    for url, fileout in zip(pharos_urls, fileouts):
        filenames.append(_download_sedm_data_from_url(
            url, fileout, return_filename, check_file, auth, verbose, overwrite, show_progress))
    return filenames if return_filename else None


#######################
#                     #
#  INTERNAL JSON DB   #
#                     #
#######################
# 20181012 20181105
EMPTY_WHAT_DF = pandas.DataFrame(columns=["filename", "airmass", "shutter",
                                          "exptime", "target", "night"])


def _parse_line_(line):
    """ """
    try:
        filename, rest = line.split('(')
        info, what = rest.split(")")
        what = what.replace(":", "")
        return [filename.replace(" ", "")]+info.split("/")+[what.replace(" [A]", "").strip()]
    except:
        return None


def whatfiles_to_dataframe(whatfile):
    """ """
    parsed_lines = [_parse_line_(l_) for l_ in whatfile]

    return pandas.DataFrame([l for l in parsed_lines if l is not None],
                            columns=["filename", "airmass", "shutter", "exptime", "target"])


def build_datelist(start="2018-06-01", end=None):
    """ """
    if end is None or end in ["today", "now"]:
        from datetime import datetime
        today = datetime.today()
        end = today.isoformat().split("T")[0]

    return ["%4d%02d%02d" % (tt.year, tt.month, tt.day)
            for tt in pandas.date_range(start=start, end=end)]


# =============== #
#                 #
#  Pharos IO      #
#                 #
# =============== #

class PharosIO(object):

    def __init__(self, whatfiles_df=None):
        """ """
        if whatfiles_df is not None:
            self.set_whatdata(whatfiles_df)

    @classmethod
    def load_local(cls, contains="*", stored=False):
        """ load instance from local whatfiles. 

        Parameters
        ----------
        contains: [string] -optional-
            get all the whatfiles containing this. 

        stored: [bool] -optional-
            load the stored multiindex DataFrame. containing
            all the whatfile loaded and then stored.
            If no whatdata has been stored, this will be ignored.

        Returns
        -------
        PharosIO instance
        """
        if stored and os.path.isfile(get_whatfile_path("stored")):
            return cls(pandas.read_parquet(get_whatfile_path("stored")))

        whatdata = cls.get_local_whatdata(contains=contains)
        return cls(whatdata)

    #
    # - get_local
    #
    @classmethod
    def get_local_whatdata(cls, contains="*", clean_exptime=True):
        """ get the multiindex dataframe concatenating all the whatfiles.

        Parameters
        ----------
        contains: [string] -optional-
            get all the whatfiles containing this. 

        clean_exptime: [bool] -optional-
            by default in the pharos whatfiles exptime entries 
            have the format '{values} s' with the ' s'. 
            If clean_exptime is True, the ' s' is removed

        Returns
        -------
        DataFrame (MultiIndex)
        """
        whatfiles = cls.get_local_whatfiles(contains)
        if len(whatfiles) == 0:
            return None

        whatdata = pandas.concat({os.path.basename(f_).split("_")[-1].split(".")[0]:
                                 pandas.read_parquet(f_)
                                 for f_ in whatfiles})

        if clean_exptime:
            whatdata["exptime"] = whatdata["exptime"].str.replace(" s", "")

        return whatdata

    @staticmethod
    def get_local_whatfiles(contains="*"):
        """ get the file of whatfiles from you computer in $ZTF/sedm/whatfiles.
        = this uses glob =

        Parameters
        ----------
        contains: [string] -optional-
            get all the whatfiles containing this. 

        Returns
        -------
        list (path)
        """
        from glob import glob
        return glob(get_whatfile_path(contains))

    @staticmethod
    def get_local_pharosfiles(contains="*"):
        """ get the file of pharosfile from you computer in $ZTF/sedm/pharosfiles.

        = this uses glob =

        Parameters
        ----------
        contains: [string] -optional-
            get all the whatfiles containing this. 

        Returns
        -------
        list (path)
        """
        from glob import glob
        return glob(get_pharosfile_path(contains))

    @staticmethod
    def _get_missingfile_dates_(which, dates):
        """ get the dates if `which` of the input dates is not in you computer.

        Parameters
        ----------
        which: [string]
            pharosfile or whatfile

        dates: [list of string]
            dates as YYYYMMDD

        Returns
        -------
        list (dates)
        """
        func = eval(f"get_{which}_path")
        return [d_ for d_ in build_datelist() if not os.path.isfile(func(d_))]

    #
    # - download
    #
    @staticmethod
    def download_whatfiles(start="2018-06-01", end=None, client=None, force_dl=False, warn=True):
        """ download the whatfiles in the given time range.

        = set client to use dask multiprocessing =

        Paramaters
        ----------
        start: [string] -optional-
            Stating date.

        end: [string or None] -optional-
            End date. 
            If None, it will be today.

        client: [dask Client]  -optional-
            Provide a dask client for using Dask multiprocessing.
            If so, a list of futures will be returned.

        force_dl: [bool] -optional-
            If the file to download already exists locally, shall this re-dlownload it ?

        warn: [bool] -optional-
            If no download are necessary, this will prompt a warning.warn,
            except if warn=False.

        Returns
        -------
        None (or list of futures if client is given)
        """
        dates = build_datelist(start=start, end=end)
        if not force_dl:
            dates = PharosIO._get_missingfile_dates_("whatfile", dates)

        if len(dates) > 0:
            return bulk_download("whatfile", np.atleast_1d(dates), client=client)
        if warn:
            warnings.warn("'whatfiles' already up to date")

    @staticmethod
    def download_pharosfiles(start="2018-06-01", end=None, client=None, force_dl=False, warn=True):
        """ download the whatfiles in the given time range.

        = set client to use dask multiprocessing =

        Paramaters
        ----------
        start: [string] -optional-
            Stating date.

        end: [string or None] -optional-
            End date. 
            If None, it will be today.

        client: [dask Client]  -optional-
            Provide a dask client for using Dask multiprocessing.
            If so, a list of futures will be returned.

        force_dl: [bool] -optional-
            If the file to download already exists locally, shall this re-dlownload it ?

        warn: [bool] -optional-
            If no download are necessary, this will prompt a warning.warn,
            except if warn=False.

        Returns
        -------
        None (or list of futures if client is given)
        """
        dates = build_datelist(start=start, end=end)
        if not force_dl:
            dates = PharosIO._get_missingfile_dates_("pharosfile", dates)

        if len(dates) > 0:
            return bulk_download("pharosfile", np.atleast_1d(dates), client=client)
        if warn:
            warnings.warn("'pharosfile' already up to date")

    @staticmethod
    def fetch_pharosfile(date, store=True, load=False, force_dl=False):
        """ """
        if load:
            store = True

        filepath = get_pharosfile_path(date)
        if force_dl or not os.path.isfile(filepath):
            _ = download_pharosfile(date, store=store)
            filepath = get_pharosfile_path(date)

        if not load:
            return filepath

        return open(filepath).read().splitlines()

    @staticmethod
    def fetch_whatfile(date, store=True, load=False, force_dl=False):
        """ """
        if load:
            store = True

        filepath = get_whatfile_path(date)
        if force_dl or not os.path.isfile(filepath):
            _ = download_whatfile(date, store=store, as_dataframe=False)
            filepath = get_whatfile_path(date)

        if not load:
            return filepath

        return pandas.read_parquet(filepath)

    #
    # - Storing
    #
    def store(self):
        """ Store locally $ZTF/sedm/whatfiles the current whatdata. """
        self.whatdata.to_parquet(get_whatfile_path("stored"))

    def update(self, force_dl=False, store=True, client=None):
        """ update the instance:
        1. call download_whatfiles() to get the missing whatfiles if any
        2. Reload all the whatfiles to rebuild whatdata.
        3. Store the updated whatdata (if store=True)

        Parameters
        ----------

        // download_whatfiles options

        force_dl: [bool] -optional-
            If the file to download already exists locally, shall this re-dlownload it ?

        client: [dask Client]  -optional-
            Provide a dask client for using Dask multiprocessing.
            If so, a list of futures will be returned.

        // other
        store: [bool] -optional-
            Shall the updated whatdata be stored (using store())


        """
        f_if_any = self.download_whatfiles(force_dl=force_dl, client=client)
        if client is not None:  # wait for the end before reloading.
            from dask.distributed import wait
            _ = wait(f_if_any)

        self.set_whatdata(self.get_local_whatdata())
        if store:
            self.store()

    # ============== #
    #  Methods       #
    # ============== #
    # -------- #
    #  SETTER  #
    # -------- #
    # -------- #
    #  SETTER  #
    # -------- #
    def set_whatdata(self, what_dataframe, infer_type=True):
        """ """
        if infer_type:
            what_dataframe = what_dataframe.astype({"airmass": "float",
                                                    "shutter": "float",
                                                    "exptime": "float"})
        self._whatdata = what_dataframe

    # -------- #
    #  GETTER  #
    # -------- #
    def get_whatdata(self, targets=None,
                     incl_calib=False, incl_std=True, incl_ztf=True,
                     calib_only=False, std_only=False, ztf_only=False,
                     airmass_range=None, exptime_range=None, date_range=None):
        """ Get the subpart of the whatdata DataFrame

        Parameters
        ----------
        targets: [string or list of] -optional-
            Select only whatdata entries corresponding to this target (or these targets).

        incl_calib, incl_std, incl_ztf: [bool] -optional-
            Include the calibration, standard star observations and ztf observations ?

        calib_only, std_only, ztf_only: [bool] -optional-
            Limit the returned entries as the one corresponding to 
            calibration, standard stars or ztf names targets.

        airmass_range, exptime_range, date_range: [float/None, float/None] -optional-
            select a range or airmass and/or exptime.
            None means no limit. For instance all the exposure lower than 300s
            will be exptime_range=[None,300]
            * Format * for date_range, dates are given as YYYYMMDD

        Returns
        -------
        DataFrame (subpart of whatdata)
        """
        def _parse_range_(din, key, krange):
            # Internal
            kmin, kmax = krange
            if kmin is None:
                return din[din[key].astype("float") < kmax]
            if kmax is None:
                return din[din[key].astype("float") > kmin]
            return din[din[key].astype("float").between(kmin, kmax)]

        data_ = self.whatdata.copy()
        if targets is not None:
            data_ = data_[data_["target"].isin(np.atleast_1d(targets))]

        if not incl_calib:
            data_ = data_[~data_["target"].str.startswith("Calib")]

        if not incl_std:
            data_ = data_[~data_["target"].str.startswith("STD")]

        if not incl_ztf:
            data_ = data_[~data_["target"].str.startswith("ZTF")]

        if std_only:
            data_ = data_[data_["target"].str.startswith("STD")]

        if ztf_only:
            data_ = data_[data_["target"].str.startswith("ZTF")]

        if calib_only:
            data_ = data_[data_["target"].str.startswith("Calib")]

        if exptime_range is not None:
            data_ = _parse_range_(data_, "exptime", exptime_range)

        if airmass_range is not None:
            data_ = _parse_range_(data_, "airmass", airmass_range)

        if date_range is not None:
            kmin, kmax = date_range
            dates = np.asarray(data_.index.get_level_values(0), dtype="float")

            if kmin is None:
                flag = dates < float(kmax)
            elif kmax is None:
                flag = dates > float(kmin)
            else:
                flag = (dates < float(kmax)) * (dates > float(kmin))
            data_ = data_[flag]

        return data_

    def get_pharosdata(self, date, force_dl=False, basename=False,
                       add_guider=True, add_snid=True):
        """ """
        if date not in self.pharosdata or force_dl:
            d = self.fetch_pharosfile(date, force_dl=force_dl, load=True)
            d += [f.replace("bkgd_", "")
                  for f in d if os.path.basename(f).startswith("bkgd_crr")]
            self.pharosdata[date] = d

        if basename:
            pdata = [os.path.basename(l) for l in self.pharosdata[date]]
        else:
            pdata = self.pharosdata[date].copy()

        if add_guider:
            targetcubes = [l for l in pdata if os.path.basename(l).startswith("e3d")
                           and "_STD-" not in l]
            guiders = [self.build_cube_astromfile(
                cubef) for cubef in targetcubes]
            pdata += guiders

        if add_snid:
            targetspecs = [l for l in pdata if os.path.basename(l).startswith("spec")
                           and "_STD-" not in l and l.endswith(".fits")]
            snidout = [self._specname_to_snidpath_(s_) for s_ in targetspecs]
            pdata += snidout

        return pdata

    def get_nights_with_target(self, targetname,
                               airmass_range=None,
                               exptime_range=None,
                               date_range=None, **kwargs):
        """ """
        target_what = self.get_whatdata(targets=targetname,
                                        airmass_range=airmass_range,
                                        exptime_range=exptime_range,
                                        date_range=date_range,
                                        **kwargs)
        return np.unique(target_what.index.get_level_values(0))

    def get_target_pharosdata(self, targetname,
                              airmass_range=None,
                              exptime_range=None,
                              date_range=None,
                              kind=None,
                              contains=None, not_contains=None,
                              extension=None,
                              add_guider=True, add_snid=True,
                              **kwargs):
        """ """
        targetbasename = self.get_target_basename(targetname=targetname,
                                                  airmass_range=airmass_range,
                                                  exptime_range=exptime_range,
                                                  date_range=date_range,
                                                  **kwargs)

        all_dates = np.unique(targetbasename.index.get_level_values(0))

        files = {}
        for date_ in all_dates:
            tfile = np.asarray(targetbasename.xs(
                date_)["basename"].values, dtype="str")
            pdata = self.get_pharosdata(date_, force_dl=False,
                                        add_guider=add_guider, add_snid=add_snid)
            files_ = [os.path.basename(l) for l in pdata if np.any([
                k in l for k in tfile])]
            # not fastest, but it's very fast anyway ; it's easier to read this way
            if not (extension is None or extension in ["*", "any", "all"]):
                files_ = [f_ for f_ in files_ if f_.endswith(extension)]

            if not (kind is None or kind in ["*", "any", "all"]):
                files_ = [f_ for f_ in files_ if f_.startswith(kind)]

            if not (contains is None or contains in ["*", "any", "all"]):
                files_ = [f_ for f_ in files_ if contains in f_]

            if not_contains is not None:
                files_ = [f_ for f_ in files_ if not not_contains in f_]

            files[date_] = files_

        return files

    def get_target_basename(self, targetname, date_range=None, **kwargs):
        """ get a dataframe the basenames assocated to the target """
        target_what = self.get_whatdata(targets=targetname,
                                        date_range=date_range,
                                        **kwargs)
        return pandas.DataFrame(target_what["filename"].astype("str"
                                                               ).str.split(".", expand=True).rename({0: "basename"}, axis=1)["basename"])

    # -------------- #
    #  Internal      #
    # -------------- #

    @classmethod
    def build_cube_astromfile(cls, cubename):
        """ """
        basename = cls._cubename_to_basename_(cubename)
        return cls.build_guiderpath(basename)

    @staticmethod
    def _cubename_to_basename_(cubename):
        """ """
        return "_".join(os.path.basename(cubename).split("_b_")[-1].split("_")[:4])

    @staticmethod
    def _specname_to_snidpath_(specname):
        """ """
        ext = specname.split(".")[-1]
        return specname.replace(".fits", "_snid.output")

    @classmethod
    def build_guiderpath(cls, basename, astrom=True):
        """ """
        date = cls.basename_to_date(basename)
        ext = "_astrom" if astrom else ""
        return os.path.join("/data", date, f"guider_crr_b_{basename}{ext}.fits.gz")

    @staticmethod
    def basename_to_date(basename):
        """ """
        return basename.split("_")[0].replace("ifu", "")

    # ============== #
    #  Properties    #
    # ============== #
    @property
    def whatdata(self):
        """ (MultiIndex) DataFrame containing the whatfiles. """
        if not hasattr(self, "_whatdata"):
            return None
        return self._whatdata

    def has_whatdata(self):
        """ """
        return self.whatdata is not None

    @property
    def pharosdata(self):
        """ dict containing the file available for a given date."""
        if not hasattr(self, "_pharosdata") or self._pharosdata is None:
            self._pharosdata = {}
        return self._pharosdata

    def has_pharosdata():
        """ """
        return len(self.pharosdata) > 0


##################
#                #
#  PHAROS        #
#                #
##################
class SEDMQuery():
    def __init__(self, load_pharosio=True, **kwargs):
        """ """
        if load_pharosio:
            self._load_pharosio_(**kwargs)

    def _load_pharosio_(self, **kwargs):
        """ """
        self._pharosio = PharosIO.load_local(stored=True, **kwargs)

    # ============= #
    #  Method       #
    # ============= #
    def update_pharosio(self, client=None, **kwargs):
        """ """
        return self.pharosio.update(client=client, **kwargs)

    def get_whatdata(self, targets=None,
                     incl_calib=False, incl_std=True, incl_ztf=True,
                     calib_only=False, std_only=False, ztf_only=False,
                     airmass_range=None, exptime_range=None, date_range=None):
        """ Get the subpart of the whatdata DataFrame

        Parameters
        ----------
        targets: [string or list of] -optional-
            Select only whatdata entries corresponding to this target (or these targets).

        incl_calib, incl_std, incl_ztf: [bool] -optional-
            Include the calibration, standard star observations and ztf observations ?

        calib_only, std_only, ztf_only: [bool] -optional-
            Limit the returned entries as the one corresponding to 
            calibration, standard stars or ztf names targets.

        airmass_range, exptime_range, date_range: [float/None, float/None] -optional-
            select a range or airmass and/or exptime.
            None means no limit. For instance all the exposure lower than 300s
            will be exptime_range=[None,300]
            * Format * for date_range, dates are given as YYYYMMDD

        Returns
        -------
        DataFrame (subpart of whatdata)
        """
        klocal = locals().copy()
        klocal.pop("self")
        return self.pharosio.get_whatdata(**klocal)

    def get_targetnames(self, **kwargs):
        """ """
        whatdata = self.get_whatdata(**kwargs)
        return np.asarray(np.unique(whatdata["target"]), dtype="str")

    # ============= #
    #  Fetch        #
    # ============= #
    # ------- #
    #  Night  #
    # ------- #
    def get_night_fluxcal(self, date,
                          download_missing=True, exist=True,
                          contains=None, not_contains=None, client=None, **kwargs):
        """ date could be a list """
        return self.get_night_data(date, kind="fluxcal", extension=".fits",
                                   download_missing=download_missing, exist=exist,
                                   contains=contains, not_contains=not_contains, client=client,
                                   **kwargs)

    def get_night_crr(self, date, contains=None, not_contains=None,
                      client=None, force_dl=False,
                      ioprop={}, **kwargs):
        """ """
        prop = dict(kind="crr", contains=contains,
                    not_contains=not_contains, extension=".fits")
        return self.get_night_data(date, client=client, ioprop=ioprop,
                                   force_dl=force_dl, **{**prop, **kwargs})

    def get_night_cubes(self, date, contains=None, not_contains="e3d_dome",
                        client=None, force_dl=False,
                        incl_dome=False,
                        ioprop={}, **kwargs):
        """ """
        prop = dict(kind="e3d", contains=contains,
                    not_contains=not_contains, extension=".fits")
        return self.get_night_data(date, client=client, ioprop=ioprop,
                                   force_dl=force_dl, **{**prop, **kwargs})

    def get_night_spectra(self, date, contains=None, not_contains=None, client=None, force_dl=False,
                          incl_dome=False, ioprop={}, **kwargs):
        """ """
        prop = dict(kind="spec", contains=contains,
                    not_contains=not_contains, extension=".fits")
        return self.get_night_data(date, client=client, ioprop=ioprop, force_dl=force_dl,
                                   **{**prop, **kwargs})

    def get_night_snidout(self, date, contains=None, not_contains=None, client=None, force_dl=False,
                          incl_dome=False, ioprop={}, **kwargs):
        """ """
        prop = dict(kind="spec", contains=contains,
                    not_contains=not_contains, extension="snid.output")
        return self.get_night_data(date, client=client, ioprop=ioprop, force_dl=force_dl,
                                   **{**prop, **kwargs})

    def get_night_astrom(self, date, not_contains=None, client=None, force_dl=False,
                         incl_dome=False, ioprop={}, **kwargs):
        """ """
        prop = dict(kind="guider", contains="astrom",
                    not_contains=not_contains, extension="*")
        return self.get_night_data(date, client=client, ioprop=ioprop, force_dl=force_dl,
                                   **{**prop, **kwargs})

    def get_night_data(self, date, download_missing=True, exist=True, force_dl=False,
                       contains=None, not_contains=None, kind=None, extension="*",
                       client=None, ioprop={}, **kwargs):
        """ """
        pharosdata = {date_:
                      [l for l in self.pharosio.get_pharosdata(date_, basename=True, **ioprop)
                       if (kind is None or l.startswith(kind)) and
                       (extension is None or extension in ["*", "all"] or
                        l.endswith(extension)) and
                       (contains is None or contains in l) and
                       (not_contains is None or not_contains not in l)]

                      for date_ in np.atleast_1d(date)}

        local_path = np.concatenate(
            self._pharosdata_to_datapath_(pharosdata, "local"))

        if download_missing or force_dl:
            index_missing = [not os.path.isfile(l) for l in local_path]
            if np.any(index_missing) or force_dl:
                if force_dl:
                    index_missing = None
                futures_ifany = self._download_pharosdata_(pharosdata, index=index_missing,
                                                           client=client, **kwargs)
                if client is not None:
                    from dask.distributed import wait
                    _ = wait(futures_ifany)

        if exist:
            return [l for l in local_path if os.path.isfile(l)]
        return local_path

    # ------- #
    # Target  #
    # ------- #
    def get_target_snidoutput(self, targetname, download_missing=True, exist=True, force_dl=False,
                              contains=None, not_contains=None, client=None, ioprop={}, **kwargs):
        """ 
        Parameters:
        -----------
        ioprop: [dict] -optional-
            used as kwargs for get_whatdata().
            Incl: incl_calib, incl_std, incl_ztf, 
                  calib_only, std_only, ztf_only,
                  airmass_range, exptime_range, date_range

        """
        prop = dict(kind="spec", contains=contains, not_contains=not_contains,
                    extension="snid.output")
        return self.get_target_datapath(targetname, client=client, ioprop=ioprop,
                                        force_dl=force_dl, **{**prop, **kwargs})

    def get_target_fluxcal(self, targetname, download_missing=True, exist=True,
                           contains=None, not_contains=None, client=None,
                           **kwargs):
        """ get nights associated to the target and downloads the fluxcal files of that night.
        uses self.get_night_fluxcal()
        """
        dates = self.pharosio.get_nights_with_target(targetname, **kwargs)
        return self.get_night_fluxcal(dates,
                                      download_missing=download_missing, exist=exist,
                                      contains=contains, not_contains=not_contains,
                                      client=client, **kwargs)

    def get_target_astrom(self, targetname, download_missing=True, exist=True, force_dl=False,
                          not_contains=None, client=None, ioprop={}, **kwargs):
        """ 
        Parameters:
        -----------
        ioprop: [dict] -optional-
            used as kwargs for get_whatdata().
            Incl: incl_calib, incl_std, incl_ztf, 
                  calib_only, std_only, ztf_only,
                  airmass_range, exptime_range, date_range

        """
        prop = dict(kind="guider", contains="astrom",
                    not_contains=not_contains, extension="*")
        return self.get_target_datapath(targetname, client=client, ioprop=ioprop,
                                        force_dl=force_dl, **{**prop, **kwargs})

    def get_target_guider(self, targetname, download_missing=True, exist=True, force_dl=False,
                          contains=None, not_contains=None, client=None, ioprop={}, **kwargs):
        """ 
        Parameters:
        -----------
        ioprop: [dict] -optional-
            used as kwargs for get_whatdata().
            Incl: incl_calib, incl_std, incl_ztf, 
                  calib_only, std_only, ztf_only,
                  airmass_range, exptime_range, date_range

        """
        prop = dict(kind="guider", contains=contains,
                    not_contains=not_contains, extension=".fits")
        return self.get_target_datapath(targetname, client=client, ioprop=ioprop,
                                        force_dl=force_dl, **{**prop, **kwargs})

    def get_target_spectra(self, targetname, download_missing=True, exist=True, force_dl=False,
                           contains=None, not_contains=None, client=None, ioprop={}, **kwargs):
        """ 
        Parameters:
        -----------
        ioprop: [dict] -optional-
            used as kwargs for get_whatdata().
            Incl: incl_calib, incl_std, incl_ztf, 
                  calib_only, std_only, ztf_only,
                  airmass_range, exptime_range, date_range

        """
        prop = dict(kind="spec", contains=contains,
                    not_contains=not_contains, extension=".fits")
        return self.get_target_datapath(targetname, client=client, ioprop=ioprop,
                                        force_dl=force_dl, **{**prop, **kwargs})

    def get_target_cubes(self, targetname, download_missing=True, exist=True, force_dl=False,
                         contains=None, not_contains=None, client=None, ioprop={},  **kwargs):
        """ 
        Parameters:
        -----------
        ioprop: [dict] -optional-
            used as kwargs for get_whatdata().
            Incl: incl_calib, incl_std, incl_ztf, 
                  calib_only, std_only, ztf_only,
                  airmass_range, exptime_range, date_range
        """
        prop = dict(kind="e3d", contains=contains,
                    not_contains=not_contains, extension=".fits")
        return self.get_target_datapath(targetname, client=client, ioprop=ioprop,
                                        force_dl=force_dl, **{**prop, **kwargs})

    def get_target_crr(self, targetname, download_missing=True, exist=True, force_dl=False,
                       contains=None, not_contains=None, client=None, ioprop={},  **kwargs):
        """ 
        Parameters:
        -----------
        ioprop: [dict] -optional-
            used as kwargs for get_whatdata().
            Incl: incl_calib, incl_std, incl_ztf, 
                  calib_only, std_only, ztf_only,
                  airmass_range, exptime_range, date_range
        """
        prop = dict(kind="crr", contains=contains,
                    not_contains=not_contains, extension=".fits")
        return self.get_target_datapath(targetname, client=client, ioprop=ioprop,
                                        force_dl=force_dl, **{**prop, **kwargs})

    def get_target_datapath(self, targetname, kind, contains=None, not_contains=None,
                            download_missing=True, exist=True, force_dl=False, verbose=False,
                            extension=".fits",  ioprop={}, client=None, **kwargs):
        """ 

        Parameters:
        -----------
        ioprop: [dict] -optional-
            used as kwargs for get_whatdata().
            Incl: incl_calib, incl_std, incl_ztf, 
                  calib_only, std_only, ztf_only,
                  airmass_range, exptime_range, date_range

        """
        prop = dict(contains=contains, not_contains=not_contains,
                    extension=extension)
        if verbose:
            print(prop)
        pharosdata = self.pharosio.get_target_pharosdata(targetname, kind=kind,
                                                         **{**prop, **ioprop})

        if pharosdata is None or len(pharosdata) == 0:
            return None

        local_path = np.concatenate(
            self._pharosdata_to_datapath_(pharosdata, "local"))
        if verbose:
            print(local_path)

        if download_missing or force_dl:
            index_missing = [not os.path.isfile(l) for l in local_path]
            if verbose:
                print(f"missing, {index_missing}")
            if np.any(index_missing) or force_dl:
                if force_dl:
                    index_missing = None

                futures_ifany = self._download_pharosdata_(pharosdata, index=index_missing,
                                                           client=client, **kwargs)
                if client is not None:
                    from dask.distributed import wait
                    _ = wait(futures_ifany)

        if exist:
            return [l for l in local_path if os.path.isfile(l)]
        return local_path

    # ============= #
    #  Download     #
    # ============= #
    # -------- #
    #  Night   #
    # -------- #
    def bulk_download_night(self, date, contains=None, not_contains=None, kind=None, extension="*",
                            dirout=None, client=None, nprocess=None, **kwargs):
        """ """
        pharosdata = {date_:
                      [l for l in self.pharosio.get_pharosdata(date_, basename=True)
                       if (kind is None or l.startswith(kind)) and
                       (extension is None or extension in ["*", "all"] or l.endswith(extension)) and
                       (contains is None or contains in l) and
                       (not_contains is None or not_contains not in l)]

                      for date_ in np.atleast_1d(date)}

        return self._download_pharosdata_(pharosdata, dirout=dirout, nprocess=nprocess, client=client,
                                          **kwargs)

    def download_night_calibrations(self, date, which="*",
                                    dirout=None, client=None, nprocess=4, **kwargs):
        """ 
        Parameters
        ----------
        date: [string (or list of)]
            night date as YYYYMMDD. It could be a list or [YYYYMMDD, YYYYMMDD etc.]

        which: [string or list of]
            kind on calibration file to download. 
            could any (of list of): 
            - ['HexaGrid.pkl', 'TraceMatch.pkl','TraceMatch_WithMasks.pkl', 
               'WaveSolution.pkl','Flatfits']
            - which='*' means all of them.

        // download options

        dirout: [string] -optional-
            where the data should be downloaded (will actually be dirout/YYYYMMDD/_downloaded_data_).
            by default this will be `$ZTDATA/sedm/redux`.

        client: [Dask Client] -optional-
            Dask client used for the downloaded. 
            If so, a list of futures will be returned.

        nprocess: [int] -optional-
            number of parallel downloading. 
            - ignored if client is given -

        **kwargs goes to download_target_data()
         incl: ignore_warnings=False, verbose=False, overwrite=True
            nodl: [bool] -optional-
                if nodl=True, nothing is downloaded and
                the list_of_url and list_of_path_where_downloaded are returned.

            show_progress: [bool] -optional-
                show the progress of downloading
                - ignored if client is used -
        """
        NIGHTLY = ["HexaGrid.pkl", "TraceMatch.pkl",
                   "TraceMatch_WithMasks.pkl", "WaveSolution.pkl", "Flat.fits"]

        if which in ["*", "all"]:
            which = NIGHTLY
        elif np.any([w_ not in NIGHTLY for w_ in np.atleast_1d(which)]):
            raise ValueError(
                f"At least one of the which input is not a known nighly calibration: {which}")

        pharosdata = {}
        for date_ in np.atleast_1d(date):
            contains = [f"{date_}_{k}" for k in np.atleast_1d(which)]
            pharosdata[date_] = [l for l in self.pharosio.get_pharosdata(date_, basename=True)
                                 if np.any([k in l for k in contains])]

        return self._download_pharosdata_(pharosdata, dirout=dirout, nprocess=nprocess, client=client,
                                          **kwargs)

    def download_night_fluxcal(self, date, contains=None, not_contains=None,
                               dirout=None, client=None, nprocess=4, **kwargs):
        """ 

        Parameters
        ----------
        date: [string (or list of)]
            night date as YYYYMMDD. It could be a list or [YYYYMMDD, YYYYMMDD etc.]

        contains: [string] -optional-
            element of the name that should be in the pharos' filename.

        not_contains: [string] -optional-
            element of the name that should *not* be in the pharos' filename.

        // download options

        dirout: [string] -optional-
            where the data should be downloaded (will actually be dirout/YYYYMMDD/_downloaded_data_).
            by default this will be `$ZTDATA/sedm/redux`.

        client: [Dask Client] -optional-
            Dask client used for the downloaded. 
            If so, a list of futures will be returned.

        nprocess: [int] -optional-
            number of parallel downloading. 
            - ignored if client is given -

        **kwargs goes to download_target_data()
         incl: ignore_warnings=False, verbose=False, overwrite=True
            nodl: [bool] -optional-
                if nodl=True, nothing is downloaded and
                the list_of_url and list_of_path_where_downloaded are returned.

            show_progress: [bool] -optional-
                show the progress of downloading
                - ignored if client is used -

        Returns
        -------

        """
        pharosdata = {date_: [l for l in self.pharosio.get_pharosdata(date_, basename=True)
                              if l.startswith("fluxcal") and
                              (contains is None or contains in l) and
                              (not_contains is None or not_contains not in l)]
                      for date_ in np.atleast_1d(date)}

        return self._download_pharosdata_(pharosdata, dirout=dirout, nprocess=nprocess, client=client,
                                          **kwargs)

    # -------- #
    #  Target  #
    # -------- #
    def download_target_astrom(self, targetname, not_contains=None,
                               dirout=None, client=None, nprocess=4, ioprop={}, **kwargs):
        """ download cubes associated to a target. 
        = uses download_target_data() with kind='e3d' =

        Parameters
        ----------
        targetname: [string]
            name to the target for which you want to download data

        contains: [string] -optional-
            element of the name that should be in the pharos' filename.

        not_contains: [string] -optional-
            element of the name that should *not* be in the pharos' filename.

        // donwload options

        dirout: [string] -optional-
            where the data should be downloaded (will actually be dirout/YYYYMMDD/_downloaded_data_).
            by default this will be `$ZTDATA/sedm/redux`.

        client: [Dask Client] -optional-
            Dask client used for the downloaded. 
            If so, a list of futures will be returned.

        nprocess: [int] -optional-
            number of parallel downloading. 
            - ignored if client is given -

        // whatdata options

        ioprop: [dict] -optional-
            used as kwargs for get_whatdata().
            Incl: incl_calib, incl_std, incl_ztf, 
                  calib_only, std_only, ztf_only,
                  airmass_range, exptime_range, date_range



        **kwargs goes to download_target_data()
         incl: nodl=False, ignore_warnings=False, show_progress=True, verbose=False, overwrite=True
            nodl: [bool] -optional-
                if nodl=True, nothing is downloaded and
                the list_of_url and list_of_path_where_downloaded are returned.

            show_progress: [bool] -optional-
                show the progress of downloading
                - ignored if client is used -s
        Returns
        -------
        None (or list of futures if client given)
        """
        prop = dict(kind="guider", contains="astrom",
                    not_contains=not_contains, extension=".fits")
        return self.download_target_data(targetname, dirout=dirout, client=client, nprocess=nprocess,
                                         ioprop=ioprop,
                                         **{**prop, **kwargs})

    def download_target_guider(self, targetname, contains=None, not_contains=None,
                               dirout=None, client=None, nprocess=4, ioprop={}, **kwargs):
        """ download cubes associated to a target. 
        = uses download_target_data() with kind='e3d' =

        Parameters
        ----------
        targetname: [string]
            name to the target for which you want to download data

        contains: [string] -optional-
            element of the name that should be in the pharos' filename.

        not_contains: [string] -optional-
            element of the name that should *not* be in the pharos' filename.

        // download options

        dirout: [string] -optional-
            where the data should be downloaded (will actually be dirout/YYYYMMDD/_downloaded_data_).
            by default this will be `$ZTDATA/sedm/redux`.

        client: [Dask Client] -optional-
            Dask client used for the downloaded. 
            If so, a list of futures will be returned.

        nprocess: [int] -optional-
            number of parallel downloading. 
            - ignored if client is given -

        // whatdata options

        ioprop: [dict] -optional-
            used as kwargs for get_whatdata().
            Incl: incl_calib, incl_std, incl_ztf, 
                  calib_only, std_only, ztf_only,
                  airmass_range, exptime_range, date_range


        **kwargs goes to download_target_data()
         incl: nodl=False, ignore_warnings=False, show_progress=True, verbose=False, overwrite=True
            nodl: [bool] -optional-
                if nodl=True, nothing is downloaded and
                the list_of_url and list_of_path_where_downloaded are returned.

            show_progress: [bool] -optional-
                show the progress of downloading
                - ignored if client is used -
        Returns
        -------
        None (or list of futures if client given)
        """
        prop = dict(kind="guider", contains=contains,
                    not_contains=not_contains, extension=".fits")
        return self.download_target_data(targetname, dirout=dirout, client=client, nprocess=nprocess,
                                         ioprop=ioprop,
                                         **{**prop, **kwargs})

    def download_target_spectra(self, targetname, contains=None, not_contains=None,
                                dirout=None, client=None, nprocess=4, ioprop={}, **kwargs):
        """ download spectra associated to a target. 
        = uses download_target_data() with kind='e3d' =

        Parameters
        ----------
        targetname: [string]
            name to the target for which you want to download data

        contains: [string] -optional-
            element of the name that should be in the pharos' filename.

        not_contains: [string] -optional-
            element of the name that should *not* be in the pharos' filename.

        // download options

        dirout: [string] -optional-
            where the data should be downloaded (will actually be dirout/YYYYMMDD/_downloaded_data_).
            by default this will be `$ZTDATA/sedm/redux`.

        client: [Dask Client] -optional-
            Dask client used for the downloaded. 
            If so, a list of futures will be returned.

        nprocess: [int] -optional-
            number of parallel downloading. 
            - ignored if client is given -

        // whatdata options

        ioprop: [dict] -optional-
            used as kwargs for get_whatdata().
            Incl: incl_calib, incl_std, incl_ztf, 
                  calib_only, std_only, ztf_only,
                  airmass_range, exptime_range, date_range


        **kwargs goes to download_target_data()
         incl: nodl=False, ignore_warnings=False, show_progress=True, verbose=False, overwrite=True
            nodl: [bool] -optional-
                if nodl=True, nothing is downloaded and
                the list_of_url and list_of_path_where_downloaded are returned.

            show_progress: [bool] -optional-
                show the progress of downloading
                - ignored if client is used -

        Returns
        -------
        None (or list of futures if client given)
        """
        prop = dict(kind="spec", contains=contains,
                    not_contains=not_contains, extension=".fits")
        return self.download_target_data(targetname, dirout=dirout, client=client, nprocess=nprocess,
                                         ioprop=ioprop,
                                         **{**prop, **kwargs})

    def download_target_cubes(self, targetname, contains=None, not_contains=None,
                              dirout=None, client=None, nprocess=4, ioprop={}, **kwargs):
        """ download cubes associated to a target. 
        = uses download_target_data() with kind='e3d' =

        Parameters
        ----------
        targetname: [string]
            name to the target for which you want to download data

        contains: [string] -optional-
            element of the name that should be in the pharos' filename.

        not_contains: [string] -optional-
            element of the name that should *not* be in the pharos' filename.

        // download options

        dirout: [string] -optional-
            where the data should be downloaded (will actually be dirout/YYYYMMDD/_downloaded_data_).
            by default this will be `$ZTDATA/sedm/redux`.

        client: [Dask Client] -optional-
            Dask client used for the downloaded. 
            If so, a list of futures will be returned.

        nprocess: [int] -optional-
            number of parallel downloading. 
            - ignored if client is given -

        // whatdata options

        ioprop: [dict] -optional-
            used as kwargs for get_whatdata().
            Incl: incl_calib, incl_std, incl_ztf, 
                  calib_only, std_only, ztf_only,
                  airmass_range, exptime_range, date_range

        **kwargs goes to download_target_data()
         incl: nodl=False, ignore_warnings=False, show_progress=True, verbose=False, overwrite=True
            nodl: [bool] -optional-
                if nodl=True, nothing is downloaded and
                the list_of_url and list_of_path_where_downloaded are returned.

            show_progress: [bool] -optional-
                show the progress of downloading
                - ignored if client is used -

        Returns
        -------
        None (or list of futures if client given)
        """
        prop = dict(kind="e3d", contains=contains,
                    not_contains=not_contains, extension=".fits")
        return self.download_target_data(targetname, dirout=dirout, client=client,
                                         nprocess=nprocess,
                                         ioprop=ioprop,
                                         **{**prop, **kwargs})

    def download_target_fluxcal(self, targetname, contains=None, not_contains=None,
                                dirout=None, client=None, nprocess=4, **kwargs):
        """ """
        dates = self.pharosio.get_nights_with_target(targetname, **kwargs)
        return self.download_night_fluxcal(dates, contains=contains, not_contains=not_contains,
                                           dirout=dirout, nprocess=nprocess, client=client,
                                           **kwargs)

    def download_target_data(self, targetname, kind,
                             contains=None, not_contains=None, extension=".fits",
                             dirout=None, nodl=False,
                             show_progress=True, nprocess=4, client=None, ioprop={}, **kwargs):
        """ 

        // whatdata options

        ioprop: [dict] -optional-
            used as kwargs for get_whatdata().
            Incl: incl_calib, incl_std, incl_ztf, 
                  calib_only, std_only, ztf_only,
                  airmass_range, exptime_range, date_range


        """
        prop = dict(contains=contains, not_contains=not_contains,
                    extension=extension)
        pharosdata = self.pharosio.get_target_pharosdata(targetname, kind=kind,
                                                         **{**prop, **ioprop})
        return self._download_pharosdata_(pharosdata,
                                          dirout=dirout, nodl=nodl,
                                          show_progress=show_progress, nprocess=nprocess,
                                          client=client,
                                          **kwargs)

    def _download_pharosdata_(self, pharosdata, dirout=None,
                              nodl=False, ignore_warnings=False,
                              show_progress=True, verbose=False, overwrite=True,
                              nprocess=4, client=None, index=None):
        """ """
        if dirout is None:
            dirout = "local"

        pharos_path = np.concatenate(
            self._pharosdata_to_datapath_(pharosdata, "pharos"))
        local_path = np.concatenate(
            self._pharosdata_to_datapath_(pharosdata, dirout))
        nprocess = np.min([len(pharos_path), nprocess])
        if index is not None:
            pharos_path = pharos_path[index]
            local_path = local_path[index]

        if nodl:
            return pharos_path, local_path

        with warnings.catch_warnings():
            if ignore_warnings:
                warnings.simplefilter("ignore")

            return io.download_url(pharos_path, local_path,
                                   show_progress=show_progress, verbose=verbose,
                                   overwrite=overwrite, nprocess=nprocess,
                                   cookies="no_cookies", client=client,
                                   filecheck=False)

    # ============= #
    #  Internal     #
    # ============= #
    @staticmethod
    def _pharosdata_to_datapath_(pharosdata, source):
        """ """
        if source == "local":
            source = os.path.join(SEDMLOCAL_BASESOURCE, "redux")
        elif source == "pharos":
            source = os.path.join(PHAROS_BASEURL, "data")
        elif source is None:
            source = ""

        return [[os.path.join(source, k, v_) for v_ in v] for k, v in pharosdata.items()]

    # ============= #
    #  Properties   #
    # ============= #
    @property
    def pharosio(self):
        """ """
        return self._pharosio
